#include "Painter.h"

//One thing you'll notice right away is my decision to hardcode absolutely every graphics
//object.  I don't use OBJ files or anything.  I did this for a few reasons...
//1) I wanted very specific objects in my game.
//2) I wanted to play around with the GL functions.
//3) I don't have any software to make OBJ files.
//4) I actually made my own filetype called KOB.  Malloc didn't like it and I basically lost 3 days
//   of work.  So...we're not doing that again.

//So now I use KLists.  They are basically Quads with defined vertices.  To see all my
//pre-defined vertices, go check out Faces.c.
//More complicated objects, I relented and used Glut spheres and cylinders.

//You'll notice that I don't use blending or check for transparencies in textures.
//I didn't go that route.  I took the HUD and particle systems as my two extras.
//Hopefully they're good enough to count for a decent pass to this project.

//I define an int to reference every texture.
TextureImage Thvrim;   //The green rim on an HV Block.
TextureImage Thvbox;   //The 4 arrows.
TextureImage Tvrim;    //The blue rim on the V Block.
TextureImage Tvbox;    //The 2 arrows pointing up/down.
TextureImage Thrim;    //Orange rim on the H Block.
TextureImage Thbox;   //Arrows pointing appropriately.
TextureImage Tdbox;   //The spider on the Death Block.
TextureImage Tdrim;   //Black rim on a death block.
klist BoxList;  //The list of predefined vertices for a simple box.
klist BoxRimList; //The list of predefined vertices for a box wireframe.

TextureImage Tbottom;  //The icy bottom to the board.
TextureImage Thole;
TextureImage Tholewall;
TextureImage Tsky;
TextureImage Twall;    //The walls for the level.
klist WallBottomList;  //The list of vertices for the bottom.
klist WallBoundsList;
klist HoleList;
klist WallNorthList;
klist WallSouthList;
klist WallEastList;  //All of these are similar.  Draw them all and you have an open box.
klist WallWestList;

TextureImage Tfury;  //My flaming textures.
TextureImage Tflame;
TextureImage Tgolem;  //The golems.  A crappy statue texture that I should probably change someday.
TextureImage Tgolemends;
klist GolemEndsList;  //The top and bottom ends to my golem.
GLUquadricObj* QuadList;  //This quadlist is used to calculate texture co-ords for my more
                          //complicated objects, like spheres and such.

TextureImage Tpidfront;
TextureImage Tpidblack;
TextureImage Tpidfeet;  //My textures for pid.
TextureImage Tpidface;
klist PidFaceList;  //My vertex list for each part of Pid.  I have different lists, because
klist PidBodyList;  //I set it so that each list may have only one texture associated with it.
klist PidFrontList;
klist PidFeetList;

TextureImage Ttrap;    //My trapmine texture.
TextureImage Ttrapedge;  //The edges sticking out of it.
klist TrapList;
klist TrapEdgeList;  //The two lists for the trapmine.

//If you haven't guessed yet, it goes on like this for awhile still.

TextureImage Tswitchbox;
TextureImage Tredswitch;
TextureImage Tyellowswitch;
klist SwitchBoxList;
klist SwitchList;

TextureImage Tcrystal;
klist CrystalList;

TextureImage Texit;
klist ExitList;

TextureImage Tarrow;
klist ArrowStemList;
klist ArrowTopList;
TextureImage Tarrowrim;
klist ArrowRimList;

TextureImage Tdirt;
TextureImage Tmirror;

bool LoadTGA(TextureImage *texture, char *filename)	// Loads A TGA File Into Memory
{    
	GLubyte		TGAheader[12]={0,0,2,0,0,0,0,0,0,0,0,0};// Uncompressed TGA Header
	GLubyte		TGAcompare[12];							// Used To Compare TGA Header
	GLubyte		header[6];						  	    // First 6 Useful Bytes From The Header
	GLuint		bytesPerPixel;							// Holds Number Of Bytes Per Pixel Used In The TGA File
	GLuint		imageSize;							// Used To Store The Image Size When Setting Aside Ram
	GLuint		temp;								// Temporary Variable
	GLuint		type=GL_RGBA;							// Set The Default GL Mode To RBGA (32 BPP)

	FILE *file = fopen(filename, "rb");						// Open The TGA File

	if(	file==NULL ||								// Does File Even Exist?
		fread(TGAcompare,1,sizeof(TGAcompare),file)!=sizeof(TGAcompare) ||	// Are There 12 Bytes To Read?
		memcmp(TGAheader,TGAcompare,sizeof(TGAheader))!=0		||	// Does The Header Match What We Want?
		fread(header,1,sizeof(header),file)!=sizeof(header))			// If So Read Next 6 Header Bytes
	{
		if (file == NULL)							// Did The File Even Exist? *Added Jim Strong*
        {
            printf("File not found.");
			return false;							// Return False
        }
		else
		{
            printf("File screwed up.");
			fclose(file);							// If Anything Failed, Close The File
			return false;							// Return False
		}
	}
	texture->width  = header[1] * 256 + header[0];					// Determine The TGA Width	(highbyte*256+lowbyte)
	texture->height = header[3] * 256 + header[2];					// Determine The TGA Height	(highbyte*256+lowbyte)
    
 	if(	texture->width	<=0	||						// Is The Width Less Than Or Equal To Zero
		texture->height	<=0	||						// Is The Height Less Than Or Equal To Zero
		(header[4]!=24 && header[4]!=32))					// Is The TGA 24 or 32 Bit?
	{
        printf("File doesn't fit header.");
		fclose(file);								// If Anything Failed, Close The File
		return false;								// Return False
	}

    texture->bpp	= header[4];							// Grab The TGA's Bits Per Pixel (24 or 32)
	bytesPerPixel	= texture->bpp/8;						// Divide By 8 To Get The Bytes Per Pixel
	imageSize	= texture->width*texture->height*bytesPerPixel;			// Calculate The Memory Required For The TGA Data

	texture->imageData=(GLubyte *)malloc(imageSize);				// Reserve Memory To Hold The TGA Data

	if(	texture->imageData==NULL ||						// Does The Storage Memory Exist?
		fread(texture->imageData, 1, imageSize, file)!=imageSize)		// Does The Image Size Match The Memory Reserved?
	{
		if(texture->imageData!=NULL)						// Was Image Data Loaded
			free(texture->imageData);					// If So, Release The Image Data

		fclose(file);								// Close The File
		return false;								// Return False
	} 

	for(GLuint i=0; i<int(imageSize); i+=bytesPerPixel)				// Loop Through The Image Data
	{										// Swaps The 1st And 3rd Bytes ('R'ed and 'B'lue)
		temp=texture->imageData[i];						// Temporarily Store The Value At Image Data 'i'
		texture->imageData[i] = texture->imageData[i + 2];			// Set The 1st Byte To The Value Of The 3rd Byte
		texture->imageData[i + 2] = temp;					// Set The 3rd Byte To The Value In 'temp' (1st Byte Value)
	}

	fclose (file);									// Close The File

	// Build A Texture From The Data
	glGenTextures(1, &texture[0].texID);						// Generate OpenGL texture IDs

	glBindTexture(GL_TEXTURE_2D, texture[0].texID);					// Bind Our Texture
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);		// Linear Filtered
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);		// Linear Filtered
	if (texture[0].bpp==24)								// Was The TGA 24 Bits
	{
   	 type=GL_RGB;								// If So Set The 'type' To GL_RGB
     gluBuild2DMipmaps( GL_TEXTURE_2D, 3, texture[0].width, texture[0].height, type, GL_UNSIGNED_BYTE, texture[0].imageData ); 	
     glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST );
     glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );     
	}
	else glTexImage2D(GL_TEXTURE_2D, 0, type, texture[0].width, texture[0].height, 0, type, GL_UNSIGNED_BYTE, texture[0].imageData);

	return true;									// Texture Building Went Ok, Return True
}

//This function paints the Klist with the texture given.
void PaintObject(klist theList, GLuint texture, GLenum Type, ColourType Col)
{
 if(Col == ISGREY)
 {
  glEnable( GL_TEXTURE_2D ); //I enable Texturing here.  I always do it here, in case it was
                             //disabled by the last object.  It's here for the worst-case scenario.
  glBindTexture( GL_TEXTURE_2D, texture);  //Draw with the texture given.
 }
 else glDisable( GL_TEXTURE_2D );
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);  //This is a default option.
  glColor3f(1, 1, 1); //If lighting doesn't affect this object, make it as bright as possible.
  glShadeModel(GL_SMOOTH);  

  if(Col == ISYELLOW)  //Colour the object based on it's colour.
    glColor4f(1, 1, 0.0, 1);
  else if(Col == ISRED) glColor4f(1, 0.0, 0.0, 1);

  //Begin with the type specified.
  glBegin(Type);
  theList = ResetList(theList);
  while(theList.Current != NULL)
  {
    //The normals are calculated and stored in the klist, so simply retrieve them.
    //My normals are per-face, not per-vertex.  I personally think per-face looks better
    //given my artstyle for the game.
    glNormal3f(theList.Current->theFace.NormalX, theList.Current->theFace.NormalY, theList.Current->theFace.NormalZ);
    //I don't automatically generate texture coordinates.  Instead I flip the textures.
    //This is because I drew my textures in Flipped RAW format.  So here I just flip them back.
    float texX = 0; float texY = 0;   
    for(int k = 0; k < theList.Current->theFace.NumVertices; k++)
     {       
       glTexCoord2f (texX,texY);
       if(texX == 0 && texY == 0)
         texX++;
       else if(texY == 0 && texX == 1)
         texY++;
       else if(texX && texY) texX--;
       glVertex3f(theList.Current->theFace.VArray[k].X, theList.Current->theFace.VArray[k].Y, -theList.Current->theFace.VArray[k].Z);
//       printf("%d %d %d", theList.Current->theFace.VArray[k].X, theList.Current->theFace.VArray[k].Y, -theList.Current->theFace.VArray[k].Z);
     }
     theList.Current = theList.Current->next;
   }       
 glEnd();
 glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);   
}

//This function loads ALL of the textures for the game.
//If you're wondering about the width/height of a raw file, you'll find it listed here.
void LoadGraphics()
{
 LoadTGA(&Thvrim,"Graphics/hvrim.tga");
 LoadTGA(&Thvbox,"Graphics/hv.tga");
 LoadTGA(&Tvrim,"Graphics/vrim.tga"); 
 LoadTGA(&Tvbox,"Graphics/v.tga"); 
 LoadTGA(&Thrim,"Graphics/hrim.tga"); 
 LoadTGA(&Thbox,"Graphics/h.tga"); 
 LoadTGA(&Tdrim,"Graphics/drim.tga"); 
 LoadTGA(&Tdbox,"Graphics/d.tga"); 
 LoadTGA(&Tcrystal,"Graphics/crystal.tga"); 
 LoadTGA(&Tarrow,"Graphics/arrow.tga");  
 LoadTGA(&Tarrowrim,"Graphics/arim.tga");  
 LoadTGA(&Texit,"Graphics/exit.tga");  
 LoadTGA(&Ttrap,"Graphics/trap.tga");  
 LoadTGA(&Ttrapedge,"Graphics/edge.tga");  
 LoadTGA(&Tswitchbox,"Graphics/switch.tga");  
 LoadTGA(&Tredswitch,"Graphics/red.tga");  
 LoadTGA(&Tyellowswitch,"Graphics/yellow.tga"); 
 LoadTGA(&Tpidfront,"Graphics/front.tga"); 
 LoadTGA(&Tpidblack,"Graphics/black.tga");  
 LoadTGA(&Tpidface,"Graphics/face.tga");  
 LoadTGA(&Tpidfeet,"Graphics/feet.tga");  
 LoadTGA(&Tfury,"Graphics/fury.tga");  
 LoadTGA(&Tflame,"Graphics/flame.tga");  
 LoadTGA(&Tgolem,"Graphics/golem.tga");  
 LoadTGA(&Tgolemends,"Graphics/golemends.tga");  
 LoadTGA(&Tbottom,"Graphics/bottom.tga");  
 LoadTGA(&Tdirt,"Graphics/dirt.tga"); 
 LoadTGA(&Tmirror, "Graphics/mirror.tga");
 LoadTGA(&Twall,"Graphics/wall.tga");  
 LoadTGA(&Tholewall, "Graphics/fade.tga");
 LoadTGA(&Thole,"Graphics/hole.tga");  

 WallNorthList = CreateWallNorth();
 WallEastList = CreateWallEast();
 WallSouthList = CreateWallSouth();
 WallWestList = CreateWallWest();
 WallBottomList = CreateWallBottom();
 WallBoundsList = CreateWallBounds();
 HoleList = CreateHole();
 GolemEndsList = CreateGolemEnds(); 

 QuadList = gluNewQuadric();			// Create a pointer to the Quadric object.
 gluQuadricNormals(QuadList, GLU_SMOOTH);	
 gluQuadricTexture(QuadList, GL_TRUE);	// It will calculate fractional co-ords for the object.
 PidFaceList = CreatePidFace();
 PidBodyList = CreatePidBody();
 PidFrontList = CreatePidFront();
 PidFeetList = CreatePidFeet();
 SwitchBoxList = CreateSwitchBox();
 SwitchList = CreateSwitch();
 ExitList = CreateExit();
 TrapList = CreateTrap();
 TrapEdgeList = CreateTrapEdge();
 ArrowStemList = CreateArrowStem();
 ArrowTopList = CreateArrowTop();
 ArrowRimList = CreateArrowRim(); 
 BoxList = CreateBox();
 BoxRimList = CreateBoxRim();
 CrystalList = CreateCrystal();
}

//The following functions are for each object.  The functions just paint all the
//klists associated with it.  If it gets more complicated than that, I'll document it.
//The colour passed is the colour for the tile we're painting.
void PaintHVBlock(ColourType Colour)
{ 
 PaintObject(BoxList, Thvbox.texID, GL_QUADS, Colour);
 PaintObject(BoxRimList, Thvrim.texID, GL_QUADS, Colour);  
}

void PaintVBlock(ColourType Colour)
{ 
 PaintObject(BoxList, Tvbox.texID, GL_QUADS, Colour);
 PaintObject(BoxRimList, Tvrim.texID, GL_QUADS, Colour);       
}

void PaintHBlock(ColourType Colour)
{
 PaintObject(BoxList, Thbox.texID, GL_QUADS, Colour);
 PaintObject(BoxRimList, Thrim.texID, GL_QUADS, Colour);   
}

void PaintDBlock(ColourType Colour)
{
 PaintObject(BoxList, Tdbox.texID, GL_QUADS, Colour);
 PaintObject(BoxRimList, Tdrim.texID, GL_QUADS, Colour);    
}

void PaintCrystal(ColourType Colour)
{
 glPushMatrix();
 PaintObject(CrystalList, Tcrystal.texID, GL_TRIANGLES, Colour);
 glPopMatrix();
 glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); 
}

void PaintArrow(ColourType Colour)
{
 PaintObject(ArrowStemList, Tarrow.texID, GL_QUADS, Colour); 
 PaintObject(ArrowTopList, Tarrow.texID, GL_TRIANGLES, Colour);
 PaintObject(ArrowRimList, Tarrowrim.texID, GL_QUADS, Colour);
}

void PaintExit(ColourType Colour)
{
 PaintObject(ExitList, Texit.texID, GL_QUADS, Colour);
}

void PaintOpenExit(ColourType Colour)
{
 PaintObject(ExitList, Texit.texID, GL_QUADS, Colour);
}

void PaintRedSwitch(ColourType Colour)
{    
 PaintObject(SwitchBoxList, Tswitchbox.texID, GL_QUADS, ISGREY);
 PaintObject(SwitchList, Tredswitch.texID, GL_TRIANGLES, ISGREY);
}

void PaintYellowSwitch(ColourType Colour)
{
 PaintObject(SwitchBoxList, Tswitchbox.texID, GL_QUADS, ISGREY);
 PaintObject(SwitchList, Tyellowswitch.texID, GL_TRIANGLES, ISGREY);
}

void PaintTrap(ColourType Colour)
{
 PaintObject(TrapList, Ttrap.texID, GL_QUADS, Colour);
 PaintObject(TrapEdgeList, Ttrapedge.texID, GL_QUADS, Colour);
}

void PaintPid(ColourType Colour)
{
  PaintObject(PidFaceList, Tpidface.texID, GL_QUADS, Colour);    
  PaintObject(PidBodyList, Tpidblack.texID, GL_QUADS, Colour);      
  PaintObject(PidFrontList, Tpidfront.texID, GL_QUADS, Colour);
  PaintObject(PidFeetList, Tpidfeet.texID, GL_QUADS, Colour);
}

void PaintAlly(ColourType Colour)
{
  PaintObject(PidFaceList, Tpidface.texID, GL_QUADS, Colour);    
  PaintObject(PidBodyList, Tpidblack.texID, GL_QUADS, Colour);     
  PaintObject(PidFrontList, Tpidfront.texID, GL_QUADS, Colour);
  PaintObject(PidFeetList, Tswitchbox.texID, GL_QUADS, Colour);     
}

//Furies and Flames use a Sphere as the base.  So I rotate the sphere based
//on the texture map and then draw it.  
void PaintFury(ColourType Colour)
{
  glEnable( GL_TEXTURE_2D );
  glBindTexture( GL_TEXTURE_2D, Tfury.texID);     
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glColor3f(0.7, 0.7, 0.7);
  glShadeModel(GL_SMOOTH);

 glPushMatrix();
 glTranslatef(0.5, 0.5, -0.5);
 glRotatef(-90, 1, 0, 0);
 glRotatef(180, 0, 1, 0);  
 gluSphere(QuadList,0.48,15,15);
 glPopMatrix();
 glPushMatrix();
 glPopMatrix();
}

void PaintFlame(ColourType Colour)
{

  glEnable( GL_TEXTURE_2D );
  glBindTexture( GL_TEXTURE_2D, Tflame.texID);     
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glColor3f(0.7, 0.7, 0.7);
  glShadeModel(GL_SMOOTH);  

 glPushMatrix();
 glTranslatef(0.5, 0.5, -0.5);
 glRotatef(-90, 1, 0, 0);
 glRotatef(180, 0, 1, 0); 
 gluSphere(QuadList,0.48,15,15);
 glPopMatrix();
 glPushMatrix();
 glPopMatrix();
}

//Golems use a cylinder between the two end pieces.
void PaintGolem(ColourType Colour)
{
 PaintObject(GolemEndsList, Tgolemends.texID, GL_QUADS, ISGREY);

  glEnable( GL_TEXTURE_2D );
  glBindTexture( GL_TEXTURE_2D, Tgolem.texID);     
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glColor3f(0.7, 0.7, 0.7);
  glShadeModel(GL_SMOOTH);

 glPushMatrix();
 glTranslatef(0.5, 0.05, -0.5); 
 glRotatef(-90, 1, 0, 0); 
 gluCylinder(QuadList, 0.48, 0.40, 0.9, 10, 10);
 glPopMatrix();
}

//This paints a string.  It'll actually account for newlines, making me pretty happy.
//I did make a few changes, but it's actually the paintstring function which I got
//from the OpenGL Lighthouse tutorial.
void PaintString(double x, double y, void* font, char* string)
{  
  char* c;

  glDisable(GL_TEXTURE_2D);

  glColor3f(1.0,1.0,1.0);
   
  glRasterPos3f(x, y, 0.9999);

   for (c = string; *c != '\0'; c++) 
   {      
    if(*c == '_') *c = ' ';
     glutBitmapCharacter(font, *c);
     if(*c == '\n') 
     {
      y = y+15;
      glRasterPos3f(x,y,0.9999); //Relax.  The bigger fonts don't use a \n, so this is fine.
     }
   }
}

void PaintInfo(char* Out)
{
 PaintString(10,20,GLUT_BITMAP_HELVETICA_12,"a,w,s,d - move the cursor");    
 PaintString(10,40,GLUT_BITMAP_HELVETICA_12,"esc - quit the level editor");    
 PaintString(10,60,GLUT_BITMAP_HELVETICA_12,"y,r,g - Turn the walls yellow, red, or grey");     
 PaintString(10,80,GLUT_BITMAP_HELVETICA_12,"Y,R,G - Turn the object yellow, red, or grey");      
 PaintString(10,100,GLUT_BITMAP_HELVETICA_12,"m - Toggle object's movement ability");       
 PaintString(300,20,GLUT_BITMAP_HELVETICA_12,"Your cursor is red if targeted object cannot move or you have no target.");        
 PaintString(10,120,GLUT_BITMAP_HELVETICA_12,"h - Put a hole in the floor");        
 PaintString(10,140,GLUT_BITMAP_HELVETICA_12,"j - Put dirt on the floor");         
 PaintString(10,160,GLUT_BITMAP_HELVETICA_12,",(<) .(>) - Delete/add walls to floor.");  

 PaintString(10,200,GLUT_BITMAP_HELVETICA_12,"`(~) - Delete object.");  
 PaintString(10,220,GLUT_BITMAP_HELVETICA_12,"1 - Create Pid.");  
 PaintString(10,240,GLUT_BITMAP_HELVETICA_12,"2 - Create Crystal.");  
 PaintString(10,260,GLUT_BITMAP_HELVETICA_12,"3 - Create Exit.");  
 PaintString(10,280,GLUT_BITMAP_HELVETICA_12,"4 - Create Ally.");  
 PaintString(10,300,GLUT_BITMAP_HELVETICA_12,"5 - Create Golem.");        
 PaintString(10,320,GLUT_BITMAP_HELVETICA_12,"6 - Create Fury.");        
 PaintString(10,340,GLUT_BITMAP_HELVETICA_12,"7 - Create Flame General.");        
 PaintString(10,360,GLUT_BITMAP_HELVETICA_12,"8 - Create Yellow Switch.");        
 PaintString(10,380,GLUT_BITMAP_HELVETICA_12,"9 - Create Red Switch.");        
 PaintString(10,400,GLUT_BITMAP_HELVETICA_12,"0 - Create Trap."); 
 PaintString(10,420,GLUT_BITMAP_HELVETICA_12,"! - Create North Arrow."); 
 PaintString(10,440,GLUT_BITMAP_HELVETICA_12,"@ - Create East Arrow."); 
 PaintString(10,460,GLUT_BITMAP_HELVETICA_12,"# - Create South Arrow."); 
 PaintString(10,480,GLUT_BITMAP_HELVETICA_12,"$ - Create West Arrow."); 
 PaintString(10,500,GLUT_BITMAP_HELVETICA_12,"% - Create HV Block."); 
 PaintString(10,520,GLUT_BITMAP_HELVETICA_12,"^ - Create V Block.");       
 PaintString(10,540,GLUT_BITMAP_HELVETICA_12,"& - Create H Block.");       
 PaintString(10,560,GLUT_BITMAP_HELVETICA_12,"* - Create Death Block.");         
 PaintString(430,560,GLUT_BITMAP_HELVETICA_12,"Press 'f' to save to ");    
 PaintString(540,560,GLUT_BITMAP_HELVETICA_12, Out);     
 
 PaintString(230,510,GLUT_BITMAP_HELVETICA_12,"You'll need to add your level to the *.wld file manually.");
 PaintString(230,530,GLUT_BITMAP_HELVETICA_12,"Give it two integers between 1-800 and 1-600 and then the desired name of your level.");     
 
 PaintString(420,580,GLUT_BITMAP_HELVETICA_18, "Kouriku Level Editor");      
}

void PaintCursor(int X, int Y, bool Move)
{
 glPushMatrix();
   glTranslatef(X, 0, Y);       
   glTranslatef(0,0.2,0);
   glScalef(0.5,0.5,0.5);
   glTranslatef(0.4, 0.0, 0.1);
   if(Move) PaintObject(WallBottomList, Tdirt.texID, GL_QUADS, ISYELLOW);         
   else  PaintObject(WallBottomList, Tdirt.texID, GL_QUADS, ISRED);         
 glPopMatrix();
}

void PaintMapBottom()
{
 glPushMatrix();
 glTranslatef(0.0, 0.1, 0.0);
  PaintObject(WallBoundsList, Tbottom.texID, GL_QUADS, ISGREY); 
 glPopMatrix();
}

void PaintHole(int x, int y, bool North, bool East, bool South, bool West)
{
 glPushMatrix();
   glTranslatef(x, 0, y);     
   PaintObject(HoleList, Thole.texID, GL_QUADS, ISGREY);

   glEnable( GL_TEXTURE_2D );     
   glBindTexture( GL_TEXTURE_2D, Tholewall.texID);      
   glEnable( GL_BLEND ); 
   glColor4f(1, 1, 1, 1);
   glBegin(GL_QUADS);
   if(North)
   {
      glTexCoord2d(0,0);
      glVertex3f(0.0, 0.1, -1.0);                
      glTexCoord2d(1,0);   
      glVertex3f(1.0, 0.1, -1.0);     
      glTexCoord2d(1,1);   
      glVertex3f(1.0, -0.6, -1.0);     
      glTexCoord2d(0,1);  
      glVertex3f(0.0, -0.6, -1.0);
   }
   if(South)
   {
      glTexCoord2d(0,0);
      glVertex3f(0.0, 0.1, 0.001);                
      glTexCoord2d(1,0);   
      glVertex3f(0.999, 0.1, 0.001);     
      glTexCoord2d(1,1);   
      glVertex3f(0.999, -0.6, 0.001);     
      glTexCoord2d(0,1);  
      glVertex3f(0.0, -0.6, 0.001);
   }
   if(West)
   {
      glTexCoord2d(0,0);
      glVertex3f(0.001, 0.1, 0.001);                
      glTexCoord2d(0,1);   
      glVertex3f(0.001, -0.6, 0.001);     
      glTexCoord2d(1,1);   
      glVertex3f(0.001, -0.6, -0.999);     
      glTexCoord2d(1,0);  
      glVertex3f(0.001, 0.1, -0.999);
   }
   if(East)
   {
      glTexCoord2d(0,0);
      glVertex3f(0.999, 0.1, 0.001);                
      glTexCoord2d(0,1);   
      glVertex3f(0.999, -0.6, 0.001);     
      glTexCoord2d(1,1);   
      glVertex3f(0.999, -0.6, -0.999);     
      glTexCoord2d(1,0);  
      glVertex3f(0.999, 0.1, -0.999);
   }         
   glEnd();
   
 glPopMatrix();      
}

//This function paints the holes in the floor.  It calculates holes connecting together
//and places walls accordingly.
void PaintHoles(eLevel theLevel)
{
  for(int i = 0; i < theLevel.Size; i++)
   for(int j = 0; j < theLevel.Size; j++)
   {
    if(theLevel.Walls[i][j].Hole)  //Check adjacent holes.  
    {
     bool North = false;
     bool West = false;  //These will be set to true if walls need to be placed there.
     bool East = false;  //ie. the pit ends at this spot for one of four directions.
     bool South = false;
     if(i-1 < 0 || !theLevel.Walls[i-1][j].Hole) West = true;
     if(i+1 >= theLevel.Size || !theLevel.Walls[i+1][j].Hole) East = true;
     if(j-1 < 0 || !theLevel.Walls[i][j-1].Hole) North = true;
     if(j+1 >= theLevel.Size || !theLevel.Walls[i][j+1].Hole) South = true;     
     PaintHole(i,j, North, East, South, West);
    }
   }
}

//This function paints the floor tiles over the reflective surface, be it dirt or the mirrory effect.
void PaintFloor(WallTile theWall, int x, int y)
{
 glPushMatrix();
   glTranslatef(x, 0, y);       
   glTranslatef(0,0.01,0);
   if(theWall.Dirty) PaintObject(WallBottomList, Tdirt.texID, GL_QUADS, ISGREY);    
   else if(!theWall.Hole) PaintObject(WallBottomList, Tmirror.texID, GL_QUADS, ISGREY);    
 glPopMatrix();
}

//This function paints walls.
void PaintWall(WallTile theWall, int x, int y)
{
 //If this is a reflection, don't paint the bottom to the level again, otherwise...that'd look weird.
 glPushMatrix();
 glTranslatef(x, 0, y);
 if(!theWall.Hole)  //If there is a hole, we just don't paint anything.
 {

   glEnable( GL_BLEND );
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); 
   PaintObject(WallBottomList, Tbottom.texID, GL_QUADS, ISGREY); //Paint the bottom. 

   //Now we paint the appropriate walls with the appropriate colour.
  if(theWall.Blocks[0]) PaintObject(WallNorthList, Twall.texID, GL_QUADS, theWall.Colour);
  if(theWall.Blocks[1]) PaintObject(WallEastList, Twall.texID, GL_QUADS, theWall.Colour);
  if(theWall.Blocks[2]) PaintObject(WallSouthList, Twall.texID, GL_QUADS, theWall.Colour); 
  if(theWall.Blocks[3]) PaintObject(WallWestList, Twall.texID, GL_QUADS, theWall.Colour);   
 }
 
 glPopMatrix();
}

//This paints the given tile.  So it takes some info from it and determines what and where
//to draw the object.

void PaintCharacter(PlayTile theTile)
{
 if(strcmp("EMPTY", theTile.Name) == 0) { return; }
 glPushMatrix();
//Then we centre it and scale it down a little.  This is so that
//the object will draw properly in-between walls.
 glTranslatef(theTile.X, 0.1, theTile.Y-0.05);  
 glScalef(0.92, 0.92, 0.92);  

 
 //Of course, the PointDirection combines with the Face direction.  The code below
 //is held exclusively by arrows at the moment.  Also, an object will have either
 //one or the other and not both.  This also could change.
 
 if(theTile.PointDirection == PEAST)
 {
  glTranslatef(0.5, 0, 0.5);
  glRotatef(-90, 0, 1, 0); 
  glTranslatef(-1.5, 0, 0.5);
 }
 else if(theTile.PointDirection == PWEST)
 {
  glTranslatef(0.5, 0, 0.5);
  glRotatef(90, 0, 1, 0);  
  glTranslatef(0.5, 0, 0.5);     
 }
 else if(theTile.PointDirection == PSOUTH)
 {
  glTranslatef(0.5, 0, 0.5);
  glRotatef(180, 0, 1, 0);
  glTranslatef(-0.5, 0, 1.5);
 }   
 
 //Use the String to determine which object to paint.
 if(strcmp("PID", theTile.Name) == 0) PaintPid(theTile.Colour);
 else if(strcmp("CRYSTAL", theTile.Name) == 0)
 {
  PaintCrystal(theTile.Colour);
 }  
 else if(strcmp("CLOSEDEXIT", theTile.Name) == 0) PaintExit(theTile.Colour);
 else if(strcmp("OPENEXIT", theTile.Name) == 0) 
 {
  PaintOpenExit(theTile.Colour); 
 }
 else if(strcmp("ALLY", theTile.Name) == 0) PaintAlly(theTile.Colour);
 else if(strcmp("GOLEM", theTile.Name) == 0) PaintGolem(theTile.Colour);
 else if(strcmp("FURY", theTile.Name) == 0) 
 {
  PaintFury(theTile.Colour);
 }
 else if(strcmp("FLAME", theTile.Name) == 0) 
 {
  PaintFlame(theTile.Colour);
 }
 else if(strcmp("SWITCH", theTile.Name) == 0) 
 {
   if(theTile.Colour == SETRED) PaintRedSwitch(theTile.Colour);
   else if(theTile.Colour == SETYELLOW) PaintYellowSwitch(theTile.Colour);
 }
 else if(strcmp("TRAP", theTile.Name) == 0) PaintTrap(theTile.Colour);
 else if(strcmp("ARROW", theTile.Name) == 0) PaintArrow(theTile.Colour);
 else if(strcmp("HVBLOCK", theTile.Name) == 0) PaintHVBlock(theTile.Colour);
 else if(strcmp("VBLOCK", theTile.Name) == 0) PaintVBlock(theTile.Colour);
 else if(strcmp("HBLOCK", theTile.Name) == 0) PaintHBlock(theTile.Colour);
 else if(strcmp("DBLOCK", theTile.Name) == 0) PaintDBlock(theTile.Colour);
 glPopMatrix(); 
}

void ExitGraphics()
{
  freeList(BoxList);
  freeList(BoxRimList); 
  freeList(WallBottomList);   
  freeList(WallNorthList);   
  freeList(WallEastList);   
  freeList(WallSouthList);   
  freeList(GolemEndsList);   
  freeList(PidFaceList);   
  freeList(PidBodyList);   
  freeList(PidFrontList);   
  freeList(PidFeetList);   
  freeList(TrapList);   
  freeList(TrapEdgeList);   
  freeList(SwitchBoxList);   
  freeList(SwitchList);   
  freeList(CrystalList);   
  freeList(ExitList);   
  freeList(ArrowTopList);   
  freeList(ArrowRimList);   
  freeList(WallBoundsList);
  freeList(HoleList);
  freeList(ArrowStemList);   
  gluDeleteQuadric(QuadList);  
}

