#include <stdio.h>
#include <stdlib.h>
#include "Map.h"


FILE* Solution;

DMarker NudgeList[10];  //This keeps track of where Nudges occured on the board in the last round.
                      //We need this to record where to play Nudge animations and of what type.
                      //It will only record up to a maximum of 10 Nudge animations per round to
                      //keep from overtaxing the CPU.

void StartRecording()
{
 Solution = fopen("temp.tmp","w");   
}

void StopRecording()
{
 fclose(Solution);    
}

//This function loads a map from a file specified.
KMap LoadMap(char* FileName)
{
  for(int i=0; i < 10; i++)  //Reset the amount of Nudges we can record for this round.
  {
    NudgeList[i].Type = -1;
    NudgeList[i].x = -1;
    NudgeList[i].y = -1;      
  }       
 int i;
 //We make a new map and set the initial flags on the board and list.
 //You'll notice that the starting active colour is ALWAYS Red.  That won't change.
 KMap theMap; 
 theMap.AnimateBusy = false;
 theMap.aLevel.Motion = false;
 theMap.aLevel.Active = ISRED;
 theMap.aLevel.Size = 10;  
 
 theMap.aLevel.NumMoves = 0;
 //Static level size could change in the future.
 theMap.PlayList = CreateNewTList(); 
 FILE* NewLevel = fopen(FileName,"r");		//This opens a File so we can read it in line by line.  

 if (!NewLevel)
 {
  //If the file doesn't exist, we quit the program immediately.
  printf("%s not found.\r\n", FileName);
  exit(1);
 }

 int LineCounter = 0;		//This will make sure we only count 10 lines into the Stage array.  So if there is any
 char nextLevelLine[27];	// more level data, it gets ignored.
 
 while (fgets(nextLevelLine,27,NewLevel)!=NULL && LineCounter < theMap.aLevel.Size)
 { 
  for(i = 0; i < theMap.aLevel.Size; i++)
  {
   //You see here that the first thing read in is a 10x10 'array' of wall tiles.  
   //I use these numbers to determine the BlockType.  These numbers are written by
   //the editor.  In other words, don't try to notepad a level because it is really confusing.
   //Trust my editor :D
   if(nextLevelLine[i] == '0') theMap.aLevel.Walls[i][LineCounter] = MakeWallTile(BNONE);
   else if(nextLevelLine[i] == '1') theMap.aLevel.Walls[i][LineCounter] = MakeWallTile(BNORTH);
   else if(nextLevelLine[i] == '2') theMap.aLevel.Walls[i][LineCounter] = MakeWallTile(BEAST);
   else if(nextLevelLine[i] == '3') theMap.aLevel.Walls[i][LineCounter] = MakeWallTile(BSOUTH);
   else if(nextLevelLine[i] == '4') theMap.aLevel.Walls[i][LineCounter] = MakeWallTile(BWEST);
   else if(nextLevelLine[i] == '5') theMap.aLevel.Walls[i][LineCounter] = MakeWallTile(BNORTHEAST);
   else if(nextLevelLine[i] == '6') theMap.aLevel.Walls[i][LineCounter] = MakeWallTile(BNORTHSOUTH);
   else if(nextLevelLine[i] == '7') theMap.aLevel.Walls[i][LineCounter] = MakeWallTile(BNORTHWEST);
   else if(nextLevelLine[i] == '8') theMap.aLevel.Walls[i][LineCounter] = MakeWallTile(BEASTSOUTH);
   else if(nextLevelLine[i] == '9') theMap.aLevel.Walls[i][LineCounter] = MakeWallTile(BEASTWEST);
   else if(nextLevelLine[i] == 'A') theMap.aLevel.Walls[i][LineCounter] = MakeWallTile(BSOUTHWEST);
   else if(nextLevelLine[i] == 'B') theMap.aLevel.Walls[i][LineCounter] = MakeWallTile(BNORTHEASTSOUTH);
   else if(nextLevelLine[i] == 'C') theMap.aLevel.Walls[i][LineCounter] = MakeWallTile(BEASTSOUTHWEST);
   else if(nextLevelLine[i] == 'D') theMap.aLevel.Walls[i][LineCounter] = MakeWallTile(BSOUTHWESTNORTH);
   else if(nextLevelLine[i] == 'E') theMap.aLevel.Walls[i][LineCounter] = MakeWallTile(BWESTNORTHEAST);
   else if(nextLevelLine[i] == 'F') theMap.aLevel.Walls[i][LineCounter] = MakeWallTile(BALL);
  }       
  LineCounter++;
//  printf("%s", nextLevelLine);	
 }
 //After we're done reading the board, we look at the list of tiles provided.
 do
 {
  if(nextLevelLine[0] == 'P') //If the next thing to be read is a play tile.
  {
    int x, y, colour, move, point;
    char* Name = (char*)malloc(512);
    char Ind;
    sscanf(nextLevelLine, "%c %d %d %d %d %d %s", &Ind, &x, &y, &colour, &move, &point, Name);
//    printf("%c %d %d %d %d %d %s\r\n", Ind, x, y, colour, move, point, Name);
    //We read in the play tile.
    ColourType Col;
    if(colour == 0) Col = ISGREY;
    else if(colour == 1) Col = ISRED;
    else Col = ISYELLOW;
    
    PointAt PAt;
    
    if(point == 0) PAt = PNONE;
    else if(point == 1) PAt = PNORTH;
    else if(point == 2) PAt = PEAST;
    else if(point == 3) PAt = PSOUTH;
    else PAt = PWEST;
    //We write down all the flags necessary and make a standard tile out of the flags which
    //can be set by the Editor.
    theMap.PlayList = addNode(theMap.PlayList, MakeStandardTile(Name, Col, move, PAt, x, y));
  }
  else if(nextLevelLine[0] == 'W') //This means we set up some wall flags instead.
  {
   int x, y, colour, dirty, hole;
   char Ind;
   sscanf(nextLevelLine, "%c %d %d %d %d %d", &Ind, &x, &y, &colour, &dirty, &hole);
 //  printf("%c %d %d %d %d %d\r\n", Ind, x, y, colour, dirty, hole);   
   
   ColourType Col;
   if(colour == 0) Col = ISGREY;
   else if(colour == 1) Col = ISRED;
   else if(colour == 2) Col = ISYELLOW;
   //We write down what and where they want the new walls done and then make the change
   //to the existing array.
   theMap.aLevel.Walls[x][y].Colour = Col;
   theMap.aLevel.Walls[x][y].Dirty = dirty;
   theMap.aLevel.Walls[x][y].Hole = hole;
  }
 }while(fgets(nextLevelLine,27,NewLevel) != NULL);
 
 fclose(NewLevel);
 //That's it, we got one new map to play with.
 return theMap;
}

//This function just tells all the tiles to move west.  Whether they do or not is up to them.
//This function does not check the walls.  This isn't done until we start running interactions.
//These move functions just set us up for the next round of interactions.
KMap MoveWest(KMap theMap)
{
 if(!theMap.aLevel.Motion)    
 {     
  theMap.PlayList = ResetList(theMap.PlayList);
  while(theMap.PlayList.Current != NULL)
  {
   theMap.PlayList.Current->theTile = SetDirectionByPlayer(theMap.PlayList.Current->theTile, WEST);  
   theMap.PlayList = nextNode(theMap.PlayList);
  }      
  theMap.aLevel.NumMoves++;
  theMap.aLevel.Motion = true;
  fprintf(Solution, "4 ");  
 }
 return theMap;
}

KMap MoveEast(KMap theMap)
{
 if(!theMap.aLevel.Motion)    
 {     
  theMap.PlayList = ResetList(theMap.PlayList);
  while(theMap.PlayList.Current != NULL)
  {
   theMap.PlayList.Current->theTile = SetDirectionByPlayer(theMap.PlayList.Current->theTile, EAST);  
   theMap.PlayList = nextNode(theMap.PlayList);
  }     
  theMap.aLevel.NumMoves++;
  theMap.aLevel.Motion = true;
  fprintf(Solution, "2 ");    
 }
 return theMap;
}

KMap MoveNorth(KMap theMap)
{
 if(!theMap.aLevel.Motion)    
 {     
  theMap.PlayList = ResetList(theMap.PlayList);
  while(theMap.PlayList.Current != NULL)
  {
   theMap.PlayList.Current->theTile = SetDirectionByPlayer(theMap.PlayList.Current->theTile, NORTH);  
   theMap.PlayList = nextNode(theMap.PlayList);
  }     
  theMap.aLevel.NumMoves++;
  theMap.aLevel.Motion = true;
  fprintf(Solution, "1 ");    
 }
 
 return theMap;
}

KMap MoveSouth(KMap theMap)
{
 if(!theMap.aLevel.Motion)    
 {
  theMap.PlayList = ResetList(theMap.PlayList);
  while(theMap.PlayList.Current != NULL)
  {
   theMap.PlayList.Current->theTile = SetDirectionByPlayer(theMap.PlayList.Current->theTile, SOUTH);  
   theMap.PlayList = nextNode(theMap.PlayList);
  }     
  theMap.aLevel.NumMoves++;
  theMap.aLevel.Motion = true;
  fprintf(Solution, "3 ");    
 }
 
 return theMap;
}

DMarker GetNudge(int MarkerNumber)
{
 if(MarkerNumber < 10) return NudgeList[MarkerNumber];
}

void RegisterNudge(int X, int Y, int Type)
{
 for(int i=0; i < 10; i++)
 {
  if(NudgeList[i].Type < 0)
  {
   NudgeList[i].Type = Type;
   NudgeList[i].x = X;
   NudgeList[i].y = Y;
   i = 10;                          
  }
 }                  
}

void TallyNudgeList(PlayTile Check1, PlayTile Check2)
{
 if(Check1.RemoveMe && ( Check1.Alignment == ENEMY || (Check1.Murderer && Check1.Suicidal)))
 {
   RegisterNudge(Check1.X, Check1.Y, 1);
 }
 if(Check2.RemoveMe && ( Check2.Alignment == ENEMY || (Check2.Murderer && Check2.Suicidal)))
 {
   RegisterNudge(Check2.X, Check2.Y, 1);      
 }
 if(Check1.RemoveMe && Check1.Collected) RegisterNudge(Check2.X, Check2.Y, 3);      
 if(Check2.RemoveMe && Check2.Collected) RegisterNudge(Check2.X, Check2.Y, 3);  
}

//This function is pretty tough.  First it determines if each tile can get past the spot it's on
//based on the walls.  Then it compares every tile to every other tile and determines if there
//are any interactions.  We order the interactions and make sure that everything happens as
//specified by the interactions given in PlayTile.  This function is easily the most complex
//and I'll do my best to explain it.

KMap Interaction(KMap theMap)
{
 if(theMap.aLevel.Motion)  //First off, is the level even moving.  This flag gets set when
 {                   //a player presses a direction key.  It's an unnecessary check really,
                     //but I do it to feel safe at night.

  for(int i=0; i < 10; i++)  //Reset the amount of Nudges we can record for this round.
  {
    NudgeList[i].Type = -1;
    NudgeList[i].x = -1;
    NudgeList[i].y = -1;      
  }  
  
  theMap.PlayList = ResetList(theMap.PlayList); //We reset the list.
  bool Stop = true; //We assume that this is the last round of interactions.
  
  while(theMap.PlayList.Current != NULL) //Run through each tile.
  {
   if(theMap.PlayList.Current->theTile.MoveType == BLOCK &&  
      theMap.PlayList.Current->theTile.Moving == STOPPED) theMap.PlayList.Current->theTile.Moving = PUSH;
   //If the tile was a Block and it was stopped for some odd reason, we set it to PUSH for the next round.  
   //This means that the Block could be pushed again in the next round.  This is to handle the possibility
   //of a block getting pushed from two different directions after it hit a wall.
   theMap.PlayList.Current->theTile = TestWalls(theMap.aLevel, theMap.PlayList.Current->theTile);
   if(theMap.PlayList.Current->theTile.RemoveMe && theMap.PlayList.Current->theTile.InTheWay) //If it ran into a hole and sizeable.
      RegisterNudge(theMap.PlayList.Current->theTile.X, theMap.PlayList.Current->theTile.Y, 2);
   //Then we test to see if the tile is still moving based on the walls it might hit.
   theMap.PlayList = nextNode(theMap.PlayList);
  }    
  
  //Now that we're done that, we check the interactions between every single Tile.  This is
  //O(n^2) and that's the end of it.  There are no possible time-savers to write here because
  //of the possibly complex interactions between tiles (ie. different directions at once, colour changes).
  TNode* Active = theMap.PlayList.Head;  
  bool Swap = false;  //We initially assume that the two tiles do not need to be swapped in this round.
  bool ChangeColour = false; //We assume that a switch will not be hit this round.  If a switch
                      //IS hit this round, we change colour after everything is done moving.  That
                      //means that if a switch is hit at the exact same time as a tile is on top of
                      //an inactive tile, there is no interaction.   And actually, if this takes place,
                      //the switch being hit could be ignored.  So it's possible that hitting a switch
                      //may not actually switch the active colour.  This is intended to avoid two blocks 
                      //that could take up space being in the same spot.

  while(Active != NULL)
  {                             
   //If a swap was made, we're looking at the new tile RIGHT now.  We need to check it again
   //for wall interactions just in case it wasn't moving before this whole thing got started.
   if(Swap) Active->theTile = TestWalls(theMap.aLevel, Active->theTile);
   //If this tile is moving in this round, then it's obviously NOT the last round.  The
   //last round is defined as all movement for all tiles being finished.  The player can
   //only move after each last round.
   if(Active->theTile.Moving != STOPPED && Active->theTile.Moving != PUSH) Stop = false;
   //We assume that the next two tiles will not need a swap.
   Swap = false;
   bool AlreadyMoved = false;  //This gets set if I end up moving the current tile before
   //I get to the end, where the tile gets to make it's move.  Sometimes I move tiles early
   //in order to make way for other important interactions.
   TNode* Reactive = theMap.PlayList.Head;
   bool TopLevelFlag = true;  //This gets us to run through the following loop twice.
   //Once for the top level tiles, and once for the rest.  To understand why I separated
   //the two possibilities, check the bottom of this function for my notes.
   while(Reactive != NULL || TopLevelFlag)
   {
    if(Reactive == NULL) //If we made it here, it's because we're done
    {   //checking the top tiles and now we need to check the rest.  
     TopLevelFlag = false;  //So we reset the list, and go through it again.
     Reactive = theMap.PlayList.Head;            
    }
    int Interact; //We don't check for an interaction if it's the same tile :p
    //We also don't check toplevel tiles during the second iteration and we
    //check only toplevel tiles during the first.
    if(Active == Reactive || (TopLevelFlag && !Reactive->theTile.InTheWay)
       || (!TopLevelFlag && Reactive->theTile.InTheWay)) Interact = 0;
    else Interact = SeeInteraction(Active->theTile, Reactive->theTile, theMap.aLevel.Active);
    //We test to see if the two have an interaction.  This function is in PlayTile.
    if(Interact == 2) //If there is no guarantee in the interaction, it means that the second
    {         //tile is moving away from the current spot.  Since it's possible that the second
              //tile could actually stop where it is (we don't know because we haven't checked IT
              //for interactions), we better just stop.  Stop and swap.
     theMap.PlayList = SwapNodes(theMap.PlayList, Active, Reactive);            
     //We swap the two tiles and just STOP checking this tile.  We'll come back to it later.
     Reactive = NULL; //End the loop.
     TopLevelFlag = false;
     Swap = true;     //This lets me know a swap just occured.  If a swap did occur, we're not
                      //actually done checking this tile, so don't let it move or anything.    
    }
    else if(Interact == 1) //If we got a guaranteed interaction.
    {
     if(Reactive->theTile.Colour == SETRED && Active->theTile.InTheWay && theMap.aLevel.Active == ISYELLOW) ChangeColour = !ChangeColour;
     else if(Reactive->theTile.Colour == SETYELLOW && Active->theTile.InTheWay && theMap.aLevel.Active == ISRED) ChangeColour = !ChangeColour;
     //If we just hit a switch, we set up the active colour to change if necessary.  We'll worry about this later.
     int result = RemoveTile(Active->theTile, Reactive->theTile, theMap.aLevel.Active);
     //We check to see what happens between the two tiles.
     if(result == 0) //If both are still around....
     {
      int direction = DetermineNewDirection(Active->theTile, Reactive->theTile, theMap.aLevel.Active);
      //Now we check to see if they have any sort of fancy effect on eachother.
      if(direction == 0) //If no fancy effect...
      {
        bool position = Takeover(Active->theTile, Reactive->theTile, theMap.aLevel.Active);
        //Check to see if they rest ontop of eachother or not.
        if(!position) 
        { 
          Active->theTile.Moving = STOPPED; Reactive->theTile.Moving = STOPPED; 
          if(Active->theTile.MoveType == BLOCK || Reactive->theTile.MoveType == BLOCK)
               RegisterNudge(0,0,5); 
        }
        //If they don't, just stop them right where they are.  We tell both to stop, because
        //we can only have an interaction if one is already stopped OR they were moving in
        //opposite directions.  We obviously wanna stop both for the second case.
      }
      else if(direction == 1) // Force a new direction on the recipient.
        Reactive->theTile.Moving = (Direction) Active->theTile.PointDirection;
      else if(direction == 2) //Force a new direction on the actor.
      {
        bool position = Takeover(Active->theTile, Reactive->theTile, theMap.aLevel.Active);
        if(!position) { Active->theTile.Moving = STOPPED; Reactive->theTile.Moving = STOPPED; }
        Active->theTile.NewMoving = (PointAt) Reactive->theTile.PointDirection;
        //The new direction will be considered next round.  For the moment, just move to where
        //you would have (ie. on top of the arrow) or stop in your tracks (a blocking arrow?  I'm accounting for custom blocks whenever I can).
      }
      else if(direction == 3) 
      {
       //This gets called when you push a block.
       Reactive->theTile = SetDirectionByBoard(Reactive->theTile, Active->theTile.Moving);
       Active->theTile.Moving = Reactive->theTile.Moving;
       if(Reactive->theTile.Moving > STOPPED) 
       //If the block just got pushed, we swap the two tiles.  We do this for a very good reason.
       //Since the second one is now moving AWAY from the first one, we need to consider it's
       //interactions before this one.  So we abandon checking this one for the moment.
       {
        theMap.PlayList = SwapNodes(theMap.PlayList, Active, Reactive); 
        Reactive = NULL;
        TopLevelFlag = false;        
        Swap = true;   
       }
      }
     }
     else if(result == 1) //We remove one of the tiles.
     {
      if(Active->theTile.Collect == COLLECTABLE && Reactive->theTile.Collect == COLLECTOR && Active->theTile.RemoveMe != true )
        Active->theTile.Collected = true;
      Active->theTile.RemoveMe = true;        
     }
     else if(result == 2) //Same here, just the other tile.
     {
       if(Reactive->theTile.Collect == COLLECTABLE && Active->theTile.Collect == COLLECTOR && Reactive->theTile.RemoveMe != true)
        Reactive->theTile.Collected = true;       
       Reactive->theTile.RemoveMe = true;        
     }
     else if(result == 3) //Both get removed.  Notice how I check for collections even here.
     { 
       Active->theTile.RemoveMe = true; 
       Reactive->theTile.RemoveMe = true; 
       if(Reactive->theTile.Collect == COLLECTABLE && Active->theTile.Collect == COLLECTOR)
         Reactive->theTile.Collected = true;  
       if(Active->theTile.Collect == COLLECTABLE && Reactive->theTile.Collect == COLLECTOR)
         Active->theTile.Collected = true;              
     }
     if(Active != NULL && Reactive != NULL) TallyNudgeList(Active->theTile, Reactive->theTile);     
     if(!Swap) Reactive = Reactive->next; //If we didn't swap, we go to the next one.
    } else if(Interact == 0) Reactive = Reactive->next; //If there was NO interaction, we never
    //even went into all that above crap.  We just look at the next one.
   } 
   if(!Swap)  //If there wasn't a swap, that means we're done checking this tile.  So we move it.
   {
     if(Active->theTile.Moving == NORTH) Active->theTile.Y--;
     else if(Active->theTile.Moving == WEST) Active->theTile.X--;
     else if(Active->theTile.Moving == EAST) Active->theTile.X++;
     else if(Active->theTile.Moving == SOUTH) Active->theTile.Y++; 
     if(Active->theTile.X < 0 || Active->theTile.X >= theMap.aLevel.Size || Active->theTile.Y < 0 || Active->theTile.Y >= theMap.aLevel.Size) 
     {
      Active->theTile.RemoveMe = true;
      if(Active->theTile.InTheWay) RegisterNudge(Active->theTile.X, Active->theTile.Y, 2);
     }
     
     if(Active->theTile.NewMoving > PNONE) //If it was given a new direction, we do it here now.
     {
      Active->theTile.Moving = (Direction) Active->theTile.NewMoving;
      Active->theTile.NewMoving = PNONE;
     }      
     Active = Active->next;
   }
  }
  theMap.PlayList = removeNodes(theMap.PlayList);  //Remove the nodes that were destroyed/collected.
  if(ChangeColour) //If the level just changed active colours, we handle it now.
  {
   ColourType Change;
   if(theMap.aLevel.Active == ISYELLOW) Change = ISRED;
   else if(theMap.aLevel.Active == ISRED) Change = ISYELLOW;    
   theMap.PlayList = ResetList(theMap.PlayList);
   TNode* Active = theMap.PlayList.Head;
   TNode* Reactive = theMap.PlayList.Head;
   bool ActualChange = true;
   while(Active != NULL)  //This whole function checks a very basic premise.
   //If we're switching colours and the things that were inactive are currently free
   //to take up space on the board (ie. nothing is blocking them from becoming active)
   //then we make the switch.  If we have Pid on top of an inactive trap, we don't make the switch.
   //That's because the Trap cannot occupy the same space as Pid.  I did this purposely.
   //What?  You want his head to a splode?  Forget it, I did it to enhance the puzzle quality of certain possible designs.
   {
    if(Active->theTile.Colour == Change)
    {
     while(Reactive != NULL)
     {
      if(Active != Reactive)
        if(Reactive->theTile.X == Active->theTile.X && Reactive->theTile.Y == Active->theTile.Y)
          if(Reactive->theTile.InTheWay && Active->theTile.InTheWay)
             ActualChange = false;
      Reactive = Reactive->next;   
     }
    }
    Active = Active->next;                    
   }
   if(ActualChange)
   {
    theMap.aLevel.Active = Change;
    RegisterNudge(0,0,4);    
   }
  }
  //If there was a stop, we stop the level.
  if(Stop) theMap.aLevel.Motion = false; 
  else theMap.AnimateBusy = true;        
  //Otherwise we set the AnimateBusy.  We'll go on to the next round after the Painter catches up.    
 }   
 
 //If you wanna know why I ran through the loop twice for toplevel tiles and the rest, this
 //is why.  It was for consistency.  My original notes follow:
      //There is one problem in this function.  The problem is when a Tile runs into two tiles 
      //simultaneously.  This can only happen if one of them isn't taking up space.  At the moment,
      //the order of interactions there is arbitrary.
      //So if a Golem is ontop of an open exit, the golem may or may not stop Pid from exiting.
      //This is a known issue, and I'll handle it when I have time.
      //In the meantime, there is a sloppy fix.  Just make sure that the exit is the last object
      //defined in the level file.  Since it'll never swap out, it's guaranteed to always be last
 //on the list.  Similar for arrows as well.  It works, but I still wanna fix it anyway.
 
 return theMap;  
}

//This function checks to see if the Painter X, Y co-ords for the given object
//have caught up with the actual X,Ys.  If not, we're still busy animating.
//We check every tile for this.
KMap CheckAnimation(KMap theMap)
{
 theMap.PlayList = ResetList(theMap.PlayList);
 theMap.AnimateBusy = false;
 while(theMap.PlayList.Current != NULL)
 {
  if(theMap.PlayList.Current->theTile.paintX != theMap.PlayList.Current->theTile.X ||
     theMap.PlayList.Current->theTile.paintY != theMap.PlayList.Current->theTile.Y) 
    {
     theMap.AnimateBusy = true;
    }
  theMap.PlayList.Current = theMap.PlayList.Current->next;                              
 }        
 
 return theMap;
}

int GetMoves(KMap theMap)
{
 return theMap.aLevel.NumMoves; 
}

bool Winner(KMap theMap)
{
 return theMap.PlayList.Winner;  
}

bool Loser(KMap theMap)
{
 return theMap.PlayList.GameOver;  
}
