#include "Painter.h"
#include "Particles.h"
#include "Picking.h"
#include "Menu.h"

//One thing you'll notice right away is my decision to hardcode absolutely every graphics
//object.  I don't use OBJ files or anything.  I did this for a few reasons...
//1) I wanted very specific objects in my game.
//2) I wanted to play around with the GL functions.
//3) I don't have any software to make OBJ files.
//4) I actually made my own filetype called KOB.  Malloc didn't like it and I basically lost 3 days
//   of work.  So...we're not doing that again.

//So now I use KLists.  They are basically Quads with defined vertices.  To see all my
//pre-defined vertices, go check out Faces.c.
//More complicated objects, I relented and used Glut spheres and cylinders.

//You'll notice that I don't use blending or check for transparencies in textures.
//I didn't go that route.  I took the HUD and particle systems as my two extras.
//Hopefully they're good enough to count for a decent pass to this project.

//I define an int to reference every texture.
TextureImage Thvrim;   //The green rim on an HV Block.
TextureImage Thvbox;   //The 4 arrows.
TextureImage Tvrim;    //The blue rim on the V Block.
TextureImage Tvbox;    //The 2 arrows pointing up/down.
TextureImage Thrim;    //Orange rim on the H Block.
TextureImage Thbox;   //Arrows pointing appropriately.
TextureImage Tdbox;   //The spider on the Death Block.
TextureImage Tdrim;   //Black rim on a death block.
klist BoxList;  //The list of predefined vertices for a simple box.
klist BoxRimList; //The list of predefined vertices for a box wireframe.

TextureImage Tlastg;  //The previous background image loaded; kept here for world changing purposes.
TextureImage Tbackg;  //The background image loaded for each world.  This is reloaded throughout gametime.
klist BackgroundList; //The one Quad used to paint it.

TextureImage Tfailure;  //Failure message.
TextureImage Tgmenu;    //In game menu.
TextureImage Thud;  //Here's the HUD.
TextureImage Ttitle;    //Title screen.
TextureImage Tsave; //save screen.
TextureImage Thelp1;    //Help screen.
TextureImage Thelp2;    //Another help screen.
TextureImage Thelp3;    //The last one.
GLuint Tmcurrent; //The current 2D screen being shown to the player.  This is only for out-of-game menus.
TextureImage Tlice;   //My ice picker for the front screen.
klist MenuList; //The quad used to render my menu screens.


TextureImage Tlnothing; //This is my icon for when no object is selected.
TextureImage Tlpenguin;  //The rest of these are icons for the selected object.
TextureImage Tlcrystal;
TextureImage Tlexit;
TextureImage Tlgolem;
TextureImage Tlfury;
TextureImage Tlswitch;
TextureImage Tltrap;
TextureImage Tlarrow;
TextureImage Tlbox;
GLuint Tlcurrent;  //The currently selected object by the picker.
klist HUDList;   //My list to render the HUD and the object selected displayed on it.

TextureImage Tbottom;  //The icy bottom to the board.
TextureImage Thole;
TextureImage Tholewall;
TextureImage Twall;    //The walls for the level.
TextureImage Tgwall;  //Used for colored walls.
klist WallBottomList;  //The list of vertices for the bottom.
klist SkyList;  //The reflection of the sky above.
TextureImage Tsky;
klist WallBoundsList;
klist HoleList;
klist WallNorthList; //I have one wall recorded and rotate it to make the others.

TextureImage Tfury;  //My flaming textures.
TextureImage Tflame;
TextureImage Tpfire;
TextureImage Tpcrystal;
TextureImage Tgolem;  //The golems.  A crappy statue texture that I should probably change someday.
TextureImage Tgolemends;
klist GolemEndsList;  //The top and bottom ends to my golem.
GLUquadricObj* QuadList;  //This quadlist is used to calculate texture co-ords for my more
                          //complicated objects, like spheres and such.

TextureImage Tworldhud;
TextureImage Twstarno;
TextureImage Twstaryes;

TextureImage Tpidfront;
TextureImage Tpidblack;
TextureImage Tpidfeet;  //My textures for pid.
TextureImage Tpidface;
klist PidFaceList;  //My vertex list for each part of Pid.  I have different lists, because
klist PidBodyList;  //I set it so that each list may have only one texture associated with it.
klist PidFrontList;
klist PidFeetList;

TextureImage Ttrap;    //My trapmine texture.
TextureImage Ttrapedge;  //The edges sticking out of it.
klist TrapList;
klist TrapEdgeList;  //The two lists for the trapmine.

//If you haven't guessed yet, it goes on like this for awhile still.

TextureImage Tswitchbox;
TextureImage Tredswitch;
TextureImage Tyellowswitch;
klist SwitchBoxList;
klist SwitchList;

TextureImage Tcrystal;
TextureImage Tgcrystal;
klist CrystalList;

TextureImage Texit;
klist ExitList;

TextureImage Tarrow;
klist ArrowStemList;
klist ArrowTopList;
TextureImage Tarrowrim;
klist ArrowRimList;

TextureImage Tcursor;
TextureImage Tsuccess;
TextureImage Tdirt;
TextureImage Tmirror;

TextureImage Texplode[3];
TextureImage Ttwinkle[3];
TextureImage Tlpid[3];
TextureImage Tlblinking[2];
int PidIconAnimate = 0;
bool ReversePid = false;
//Okay, here we go.
char* Info = "";  //This string is what is currently displayed in the info box.
int ColorCode = 0; //This is an indicator used to determine if we're in picking mode or not.
                // The colors used to draw objects will vary depending on the answer.
ColourType Active = ISRED;  //Sets the active colour for the level.  Always starts on Red.
bool Skidmark = false;
bool CrystalRotate = false;
float RotateCrystal = 0.0;
bool Shifting = false;  //When the map is changing on the worldscreen, we shift right in our special effect.  This lets us know it's still going on.
int MaxX = 800;
int ShiftX = 0;

void ShiftMap()
{  
 Shifting = true;
 if(ShiftingLeft())
 {  
  ShiftX = MaxX;
 }
 else ShiftX = -MaxX;
}

bool StillShifting()
{
 return Shifting;    
}

Animation AnimList[10];  //Only allow 10 animations on-screen at any time.  
//There is a function to overwrite the oldest animations to make way for newer
//ones, whether the old ones are done or not.  

int IconCounter = 0;
void AnimatePidIcon(bool Anim)
{
 if(Anim) IconCounter++;
 if(IconCounter > 20)
 {
  IconCounter = 0;
  if(!ReversePid)
  {   
   PidIconAnimate++;
   if(PidIconAnimate > 2) 
   {
    PidIconAnimate = 1;
    ReversePid = !ReversePid;
   }
  }
  else
  {
   PidIconAnimate--;
   if(PidIconAnimate < 0) 
   {
    PidIconAnimate = 1;
    ReversePid = !ReversePid;
   }      
  }
 }   
}

void PaintAnimation(int pos)
{
 float modelview[16];     
 glPushMatrix();
   glTranslatef(AnimList[pos].x, 0.3, AnimList[pos].y-0.3);
   if(AnimList[pos].Type == 1) glBindTexture( GL_TEXTURE_2D, Texplode[AnimList[pos].Frame - 1].texID); 
   else if(AnimList[pos].Type == 2) glBindTexture( GL_TEXTURE_2D, Ttwinkle[AnimList[pos].Frame - 1].texID); 
   glGetFloatv(GL_MODELVIEW_MATRIX , modelview);

     for(int i=0; i<3; i++ ) 
    	for(int j=0; j<3; j++ ) 
        {
    		if ( i==j )
    			modelview[i*4+j] = 1.0;
    		else
	    		modelview[i*4+j] = 0.0;
    	}
     
   glLoadMatrixf(modelview);        
   glColor4f(1, 1, 1, 1);
   glBegin(GL_QUADS);
     glTexCoord2d(0,1);
     glVertex3f(0, 0, 0);                
     glTexCoord2d(1,1);   
     glVertex3f(0.5, 0, 0);     
     glTexCoord2d(1,0);   
     glVertex3f(0.5, 0.5, 0);     
     glTexCoord2d(0,0);  
     glVertex3f(0, 0.5, 0);     
   glEnd();
 glPopMatrix();
}

void ReduceAnimations()
{
 for(int i=0; i < 10; i++)
 {
   if(AnimList[i].Frame > 0) AnimList[i].Frame--;  
 }    
}

int Frames = 0;
void PaintAnimations(bool Anim)
{
  if(Anim) Frames++;
  if(Frames > 10) 
  {
    ReduceAnimations();
    Frames = 0;
  } 
  bool AnimationsToday = false;
  for(int i=0; i < 10; i++)
  {
     if(AnimList[i].Frame > 0) AnimationsToday = true;
  }
  if(AnimationsToday)
  {
   glEnable( GL_TEXTURE_2D );         
   glEnable( GL_BLEND );
   glDisable( GL_LIGHTING );
   for(int j=0; j < 10; j++)
   {
     for(int i=0; i<10; i++)
     {
      if(AnimList[i].Frame > 0 && AnimList[i].y == j) PaintAnimation(i);
      else if(AnimList[i].Frame == 0) i = 10;
     }
   }
     
   glEnable(GL_LIGHTING);
  }  
}

void AddAnimation(int Type, int x, int y)
{
 //First find the oldest animation in the list.
 for(int j=0; j < 3; j++)
 {
  for(int i=0; i < 10; i++)
  {
   if(AnimList[i].Frame == j)
   {
    AnimList[i].Frame = 3;
    AnimList[i].x = x;
    AnimList[i].y = y;
    AnimList[i].Type = Type;
    i = 10;
    j = 3;
   }        
  }    
 }
}

void AddExplosion(int x, int y)
{
 AddAnimation(1, x, y);
}

void AddTwinkle(int x, int y)
{
 AddAnimation(2, x, y);    
}

bool LoadTGA(TextureImage *texture, char *filename)	// Loads A TGA File Into Memory
{    
	GLubyte		TGAheader[12]={0,0,2,0,0,0,0,0,0,0,0,0};// Uncompressed TGA Header
	GLubyte		TGAcompare[12];							// Used To Compare TGA Header
	GLubyte		header[6];						  	    // First 6 Useful Bytes From The Header
	GLuint		bytesPerPixel;							// Holds Number Of Bytes Per Pixel Used In The TGA File
	GLuint		imageSize;							// Used To Store The Image Size When Setting Aside Ram
	GLuint		temp;								// Temporary Variable
	GLuint		type=GL_RGBA;							// Set The Default GL Mode To RBGA (32 BPP)

	FILE *file = fopen(filename, "rb");						// Open The TGA File

	if(	file==NULL ||								// Does File Even Exist?
		fread(TGAcompare,1,sizeof(TGAcompare),file)!=sizeof(TGAcompare) ||	// Are There 12 Bytes To Read?
		memcmp(TGAheader,TGAcompare,sizeof(TGAheader))!=0		||	// Does The Header Match What We Want?
		fread(header,1,sizeof(header),file)!=sizeof(header))			// If So Read Next 6 Header Bytes
	{
		if (file == NULL)							// Did The File Even Exist? *Added Jim Strong*
        {
            printf("File not found.");
			return false;							// Return False
        }
		else
		{
            printf("File screwed up.");
			fclose(file);							// If Anything Failed, Close The File
			return false;							// Return False
		}
	}
	texture->width  = header[1] * 256 + header[0];					// Determine The TGA Width	(highbyte*256+lowbyte)
	texture->height = header[3] * 256 + header[2];					// Determine The TGA Height	(highbyte*256+lowbyte)
    
 	if(	texture->width	<=0	||						// Is The Width Less Than Or Equal To Zero
		texture->height	<=0	||						// Is The Height Less Than Or Equal To Zero
		(header[4]!=24 && header[4]!=32))					// Is The TGA 24 or 32 Bit?
	{
        printf("File doesn't fit header.");
		fclose(file);								// If Anything Failed, Close The File
		return false;								// Return False
	}

    texture->bpp	= header[4];							// Grab The TGA's Bits Per Pixel (24 or 32)
	bytesPerPixel	= texture->bpp/8;						// Divide By 8 To Get The Bytes Per Pixel
	imageSize	= texture->width*texture->height*bytesPerPixel;			// Calculate The Memory Required For The TGA Data

	texture->imageData=(GLubyte *)malloc(imageSize);				// Reserve Memory To Hold The TGA Data

	if(	texture->imageData==NULL ||						// Does The Storage Memory Exist?
		fread(texture->imageData, 1, imageSize, file)!=imageSize)		// Does The Image Size Match The Memory Reserved?
	{
		if(texture->imageData!=NULL)						// Was Image Data Loaded
			free(texture->imageData);					// If So, Release The Image Data

		fclose(file);								// Close The File
		return false;								// Return False
	} 

	for(GLuint i=0; i<int(imageSize); i+=bytesPerPixel)				// Loop Through The Image Data
	{										// Swaps The 1st And 3rd Bytes ('R'ed and 'B'lue)
		temp=texture->imageData[i];						// Temporarily Store The Value At Image Data 'i'
		texture->imageData[i] = texture->imageData[i + 2];			// Set The 1st Byte To The Value Of The 3rd Byte
		texture->imageData[i + 2] = temp;					// Set The 3rd Byte To The Value In 'temp' (1st Byte Value)
	}

	fclose (file);									// Close The File

	// Build A Texture From The Data
	glGenTextures(1, &texture[0].texID);						// Generate OpenGL texture IDs

	glBindTexture(GL_TEXTURE_2D, texture[0].texID);					// Bind Our Texture
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);		// Linear Filtered
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);		// Linear Filtered
	if (texture[0].bpp==24)								// Was The TGA 24 Bits
	{
   	 type=GL_RGB;								// If So Set The 'type' To GL_RGB
     gluBuild2DMipmaps( GL_TEXTURE_2D, 3, texture[0].width, texture[0].height, type, GL_UNSIGNED_BYTE, texture[0].imageData ); 	
     glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST );
     glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );     
	}
	else glTexImage2D(GL_TEXTURE_2D, 0, type, texture[0].width, texture[0].height, 0, type, GL_UNSIGNED_BYTE, texture[0].imageData);

	return true;									// Texture Building Went Ok, Return True
}

//This function paints the Klist with the texture given.
void PaintObject(klist theList, GLuint texture, GLenum Type, ColourType Col, bool NoLighting, double Alpha)
{
 if(ColorCode == 1) //If we're in picking mode...
 {
  glDisable( GL_TEXTURE_2D ); //Disable texturing and draw a gray representation of the object.
  glColor3ub(theList.ColorCode, theList.ColorCode, theList.ColorCode);    //The gray representation
  //of the object is stored in the klist associated with it.     
 }
 else
 {
  glEnable( GL_TEXTURE_2D ); //I enable Texturing here.  I always do it here, in case it was
                             //disabled by the last object.  It's here for the worst-case scenario.
  glBindTexture( GL_TEXTURE_2D, texture);  //Draw with the texture given.
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);  //This is a default option.
  if(NoLighting) glColor4f(1, 1, 1, Alpha); //If lighting doesn't affect this object, make it as bright as possible.
  else glColor4f(0.7, 0.7, 0.7, Alpha); //Otherwise shade it slightly so that lighting can still affect it.
  glShadeModel(GL_SMOOTH);  
  float NewAlpha = 1.0;
   if(Col != Active && Col != ISGREY) //If this object is currently inactive...
   {
     glBlendFunc(GL_ONE, GL_ONE);
     glDisable( GL_TEXTURE_2D );  //Turn off the textures.
     NewAlpha -= 0.5;
   }
   if(Col == ISYELLOW)  //Colour the object based on it's colour.
     glColor4f(1, 1, 0.0, NewAlpha);
   else if(Col == ISRED) glColor4f(1, 0.0, 0.0, NewAlpha);
  }
  if(Col != ISGREY) glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, GL_ADD);  //This is a default option.
  //Begin with the type specified.
  glBegin(Type);
  theList = ResetList(theList);
  while(theList.Current != NULL)
  {
    //The normals are calculated and stored in the klist, so simply retrieve them.
    //My normals are per-face, not per-vertex.  I personally think per-face looks better
    //given my artstyle for the game.
    glNormal3f(theList.Current->theFace.NormalX, theList.Current->theFace.NormalY, theList.Current->theFace.NormalZ);
    //I don't automatically generate texture coordinates.  Instead I flip the textures.
    //This is because I drew my textures in Flipped RAW format.  So here I just flip them back.
    float texX = 0; float texY = 0;   
    for(int k = 0; k < theList.Current->theFace.NumVertices; k++)
     {       
       glTexCoord2f (texX,texY);
       if(texX == 0 && texY == 0)
         texX++;
       else if(texY == 0 && texX == 1)
         texY++;
       else if(texX && texY) texX--;
       glVertex3f(theList.Current->theFace.VArray[k].X, theList.Current->theFace.VArray[k].Y, -theList.Current->theFace.VArray[k].Z);
//       printf("%d %d %d", theList.Current->theFace.VArray[k].X, theList.Current->theFace.VArray[k].Y, -theList.Current->theFace.VArray[k].Z);
     }
     theList.Current = theList.Current->next;
   }       
 glEnd();
 glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);   
}

void LoadWorld(char* backg, char* sky, char* wall)
{
 LoadTGA(&Twall, wall);  
 Tlastg = Tbackg;
 LoadTGA(&Tbackg, backg);  
 LoadTGA(&Tsky, sky);      
}

void ReloadWorld()
{
 LoadWorld(GetWorldBackground(), GetWorldSky(), GetWorldWall());     
}

//This function loads ALL of the textures for the game.
void LoadGraphics()
{
 LoadWorld(GetWorldBackground(), GetWorldSky(), GetWorldWall());
 
 for(int i = 0; i < 10; i++)
 {
  AnimList[i].Frame = 0;
  AnimList[i].x = 0;
  AnimList[i].y = 0;
  AnimList[i].Type = 0;        
 } 
 LoadTGA(&Texplode[2], "Graphics/explode1.tga");
 LoadTGA(&Texplode[1], "Graphics/explode2.tga");
 LoadTGA(&Texplode[0], "Graphics/explode3.tga");  

 LoadTGA(&Ttwinkle[2], "Graphics/twinkle1.tga");
 LoadTGA(&Ttwinkle[1], "Graphics/twinkle2.tga");
 LoadTGA(&Ttwinkle[0], "Graphics/twinkle3.tga");   

 LoadTGA(&Tlpid[0],"Graphics/lpid2.tga"); 
 LoadTGA(&Tlpid[1],"Graphics/lpid.tga"); 
 LoadTGA(&Tlpid[2],"Graphics/lpid3.tga");  
 
 LoadTGA(&Tlblinking[0],"Graphics/lnext.tga");  
  LoadTGA(&Tlblinking[1],"Graphics/lnext2.tga");  

 LoadTGA(&Twstarno,"Graphics/nostar.tga"); 
 LoadTGA(&Twstaryes,"Graphics/star.tga");   
 
 LoadTGA(&Thvrim,"Graphics/hvrim.tga");
 LoadTGA(&Thvbox,"Graphics/hv.tga");
 LoadTGA(&Tvrim,"Graphics/vrim.tga"); 
 LoadTGA(&Tvbox,"Graphics/v.tga"); 
 LoadTGA(&Thrim,"Graphics/hrim.tga"); 
 LoadTGA(&Thbox,"Graphics/h.tga"); 
 LoadTGA(&Tdrim,"Graphics/drim.tga"); 
 LoadTGA(&Tdbox,"Graphics/d.tga"); 
 LoadTGA(&Tcrystal,"Graphics/crystal.tga"); 
 LoadTGA(&Tarrow,"Graphics/arrow.tga");  
 LoadTGA(&Tarrowrim,"Graphics/arim.tga");  
 LoadTGA(&Texit,"Graphics/exit.tga");  
 LoadTGA(&Ttrap,"Graphics/trap.tga");  
 LoadTGA(&Ttrapedge,"Graphics/edge.tga");  
 LoadTGA(&Tswitchbox,"Graphics/switch.tga");  
 LoadTGA(&Tredswitch,"Graphics/red.tga");  
 LoadTGA(&Tyellowswitch,"Graphics/yellow.tga"); 
 LoadTGA(&Tpidfront,"Graphics/front.tga"); 
 LoadTGA(&Tpidblack,"Graphics/black.tga");  
 LoadTGA(&Tpidface,"Graphics/face.tga");  
 LoadTGA(&Tpidfeet,"Graphics/feet.tga");  
 LoadTGA(&Tfury,"Graphics/fury.tga");  
 LoadTGA(&Tflame,"Graphics/flame.tga");  
 LoadTGA(&Tgolem,"Graphics/golem.tga");  
 LoadTGA(&Tgolemends,"Graphics/golemends.tga");  
 LoadTGA(&Tbottom,"Graphics/bottom.tga");  
 LoadTGA(&Tdirt,"Graphics/dirt.tga"); 
 LoadTGA(&Tmirror, "Graphics/mirror.tga");
 LoadTGA(&Tgwall, "Graphics/gwall.tga");
 LoadTGA(&Tgcrystal, "Graphics/gcrystal.tga");
 LoadTGA(&Tlnothing,"Graphics/lnothing.tga");  
 LoadTGA(&Tlcrystal,"Graphics/lcrystal.tga");  
 LoadTGA(&Tlpenguin,"Graphics/lpenguin.tga");  
 LoadTGA(&Tlexit,"Graphics/lexit.tga");  
 LoadTGA(&Tlgolem,"Graphics/lgolem.tga");  
 LoadTGA(&Tlfury,"Graphics/lfury.tga");  
 LoadTGA(&Tlswitch,"Graphics/lswitch.tga");  
 LoadTGA(&Tltrap,"Graphics/ltrap.tga");  
 LoadTGA(&Tlarrow,"Graphics/larrow.tga");  
 LoadTGA(&Tlbox,"Graphics/lbox.tga");  
 LoadTGA(&Ttitle,"Graphics/title.tga");   
 LoadTGA(&Tsave,"Graphics/save.tga");
 LoadTGA(&Thelp1,"Graphics/Help1.tga");  
 LoadTGA(&Thelp2,"Graphics/Help2.tga");   
 LoadTGA(&Thelp3,"Graphics/Help3.tga");  
 LoadTGA(&Tsuccess,"Graphics/success.tga");
 LoadTGA(&Thud,"Graphics/hud.tga");
 LoadTGA(&Tfailure,"Graphics/fail.tga");
 LoadTGA(&Tgmenu,"Graphics/gmenu.tga");
 LoadTGA(&Tholewall, "Graphics/fade.tga");
 Tmcurrent = Ttitle.texID;
 LoadTGA(&Tworldhud, "Graphics/worldhud.tga"); 
 LoadTGA(&Tgmenu,"Graphics/gmenu.tga");
 LoadTGA(&Tlice,"Graphics/lice.tga"); 
 LoadTGA(&Tpfire,"Graphics/pfire.tga"); 
 LoadTGA(&Tpcrystal,"Graphics/pcrystal.tga");  
 InitializeSystems(Tpfire.texID, Tpcrystal.texID);
 LoadTGA(&Thole,"Graphics/hole.tga");  
 LoadTGA(&Tcursor,"Graphics/cursor.tga");
 MenuList = CreateMenus();
 Tlcurrent = Tlnothing.texID;
//Set up the particle systems.
 InitializeFire();
 InitializeExit();
 InitializeLights();
 InitializeSkidMark();
//Set up the Klists for each object and then assign it a color code for picking purposes. 
 HUDList = CreateHUD();   
 HUDList.ColorCode = 0;
 BackgroundList = CreateBackground();
 BackgroundList.ColorCode = 0;
 WallNorthList = CreateWallNorth();
 WallNorthList.ColorCode = 0;
 WallBottomList = CreateWallBottom();
 WallBottomList.ColorCode = 0; 
 WallBoundsList = CreateWallBounds();
 WallBoundsList.ColorCode = 0;
 SkyList = CreateSky();
 WallBoundsList.ColorCode = 0;
 HoleList = CreateHole();
 HoleList.ColorCode = 0; 
 GolemEndsList = CreateGolemEnds(); 
 GolemEndsList.ColorCode = PGOLEM;
 QuadList = gluNewQuadric();			// Create a pointer to the Quadric object.
 gluQuadricNormals(QuadList, GLU_SMOOTH);	
 gluQuadricTexture(QuadList, GL_TRUE);	// It will calculate fractional co-ords for the object.
 PidFaceList = CreatePidFace();
 PidFaceList.ColorCode = PPID;
 PidBodyList = CreatePidBody();
 PidBodyList.ColorCode = PPID;
 PidFrontList = CreatePidFront();
 PidFrontList.ColorCode = PPID;
 PidFeetList = CreatePidFeet();
 PidFeetList.ColorCode = PPID;
 SwitchBoxList = CreateSwitchBox();
 SwitchBoxList.ColorCode = PSWITCH;
 SwitchList = CreateSwitch();
 SwitchList.ColorCode = PSWITCH; 
 ExitList = CreateExit();
 ExitList.ColorCode = PEXIT; 
 TrapList = CreateTrap();
 TrapList.ColorCode = PTRAP;
 TrapEdgeList = CreateTrapEdge();
 TrapEdgeList.ColorCode = PTRAP;
 ArrowStemList = CreateArrowStem();
 ArrowStemList.ColorCode = PARROW;
 ArrowTopList = CreateArrowTop();
 ArrowTopList.ColorCode = PARROW; 
 ArrowRimList = CreateArrowRim(); 
 ArrowRimList.ColorCode = PARROW; 
 BoxList = CreateBox();
 BoxList.ColorCode = PBOX; 
 BoxRimList = CreateBoxRim();
 BoxRimList.ColorCode = PBOX;  
 CrystalList = CreateCrystal();
 CrystalList.ColorCode = PCRYSTAL; 
}

//The following functions are for each object.  The functions just paint all the
//klists associated with it.  If it gets more complicated than that, I'll document it.
//The colour passed is the colour for the tile we're painting.
void PaintHVBlock(ColourType Colour)
{ 
 PaintObject(BoxList, Thvbox.texID, GL_QUADS, Colour, false, 1.0);
 PaintObject(BoxRimList, Thvrim.texID, GL_QUADS, Colour, false, 1.0);  
}

void PaintVBlock(ColourType Colour)
{ 
 PaintObject(BoxList, Tvbox.texID, GL_QUADS, Colour, false, 1.0);
 PaintObject(BoxRimList, Tvrim.texID, GL_QUADS, Colour, false, 1.0);       
}

void PaintHBlock(ColourType Colour)
{
 PaintObject(BoxList, Thbox.texID, GL_QUADS, Colour, false, 1.0);
 PaintObject(BoxRimList, Thrim.texID, GL_QUADS, Colour, false, 1.0);   
}

void PaintDBlock(ColourType Colour)
{
 PaintObject(BoxList, Tdbox.texID, GL_QUADS, Colour, false, 1.0);
 PaintObject(BoxRimList, Tdrim.texID, GL_QUADS, Colour, false, 1.0);    
}

void PaintCrystal(ColourType Colour, bool Anim)
{
 if(ColorCode != 1) glBlendFunc(GL_SRC_COLOR, GL_ONE_MINUS_SRC_ALPHA); 
 glPushMatrix();
 glTranslatef(0.5, 0.5, -0.5);
 if(Anim && CrystalRotate)  
 {
  RotateCrystal += 0.8;
  CrystalRotate = false;
 }else if (!Anim) CrystalRotate = true;
 
 if(RotateCrystal > 360) RotateCrystal = 0;
 glRotatef(RotateCrystal, 0, 1, 0);
 glTranslatef(-0.5, -0.5, 0.5);
 if(Colour == ISGREY) PaintObject(CrystalList, Tcrystal.texID, GL_TRIANGLES, Colour, true, 0.7);
 else  PaintObject(CrystalList, Tgcrystal.texID, GL_TRIANGLES, Colour, true, 0.7);
 glPopMatrix();
 glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); 
}

void PaintArrow(ColourType Colour)
{
 PaintObject(ArrowStemList, Tarrow.texID, GL_QUADS, Colour, false, 1.0); 
 PaintObject(ArrowTopList, Tarrow.texID, GL_TRIANGLES, Colour, false, 1.0);
 PaintObject(ArrowRimList, Tarrowrim.texID, GL_QUADS, Colour, false, 1.0);
}

void PaintExit(ColourType Colour)
{
 PaintObject(ExitList, Texit.texID, GL_QUADS, Colour, false, 1.0);
}

void PaintOpenExit(ColourType Colour)
{
 PaintObject(ExitList, Texit.texID, GL_QUADS, Colour, false, 1.0);
}

void PaintRedSwitch(ColourType Colour)
{    
 PaintObject(SwitchBoxList, Tswitchbox.texID, GL_QUADS, ISGREY, false, 1.0);
 PaintObject(SwitchList, Tredswitch.texID, GL_TRIANGLES, ISGREY, false, 1.0);
}

void PaintYellowSwitch(ColourType Colour)
{
 PaintObject(SwitchBoxList, Tswitchbox.texID, GL_QUADS, ISGREY, false, 1.0);
 PaintObject(SwitchList, Tyellowswitch.texID, GL_TRIANGLES, ISGREY, false, 1.0);
}

void PaintTrap(ColourType Colour)
{
 PaintObject(TrapList, Ttrap.texID, GL_QUADS, Colour, false, 1.0);
 PaintObject(TrapEdgeList, Ttrapedge.texID, GL_QUADS, Colour, false, 1.0);
}

void PaintPid(ColourType Colour, bool Anim)
{
  PaintObject(PidFaceList, Tpidface.texID, GL_QUADS, Colour, false, 1.0);    
  PaintObject(PidBodyList, Tpidblack.texID, GL_QUADS, Colour, false, 1.0);      
  PaintObject(PidFrontList, Tpidfront.texID, GL_QUADS, Colour, false, 1.0);
  PaintObject(PidFeetList, Tpidfeet.texID, GL_QUADS, Colour, false, 1.0);
  if(Skidmark)
   Skidmark = PaintSkidMark(Anim);
}

void PaintAlly(ColourType Colour)
{
  PaintObject(PidFaceList, Tpidface.texID, GL_QUADS, Colour, false, 1.0);    
  PaintObject(PidBodyList, Tpidblack.texID, GL_QUADS, Colour, false, 1.0);     
  PaintObject(PidFrontList, Tpidfront.texID, GL_QUADS, Colour, false, 1.0);
  PaintObject(PidFeetList, Tswitchbox.texID, GL_QUADS, Colour, false, 1.0);     
}

//Furies and Flames use a Sphere as the base.  So I rotate the sphere based
//on the texture map and then draw it.  
void PaintFury(ColourType Colour)
{
 if(ColorCode == 1) 
 {
  glDisable( GL_TEXTURE_2D);                 
  glColor3ub(PFURY, PFURY, PFURY);     
 }
 else
 {
  glEnable( GL_TEXTURE_2D );
  glBindTexture( GL_TEXTURE_2D, Tfury.texID);     
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glColor3f(1, 1, 1);
  glShadeModel(GL_SMOOTH);
 }
 glPushMatrix();
 glTranslatef(0.5, 0.5, -0.5);
 glRotatef(-90, 1, 0, 0);
 glRotatef(180, 0, 1, 0);  
 glDisable( GL_LIGHTING );
 gluSphere(QuadList,0.48,14,14);
 glEnable( GL_LIGHTING );
 glPopMatrix();
 glPushMatrix();
 glPopMatrix();
}

void PaintFlame(ColourType Colour)
{
 if(ColorCode == 1) 
 {
  glDisable( GL_TEXTURE_2D);              
  glColor3ub(PFLAME, PFLAME, PFLAME);
 }
 else
 {
  glEnable( GL_TEXTURE_2D );
  glBindTexture( GL_TEXTURE_2D, Tflame.texID);     
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glColor3f(1, 1, 1);
  glShadeModel(GL_SMOOTH);  
 }
 
 glPushMatrix();
 glTranslatef(0.5, 0.5, -0.5);
 glRotatef(-90, 1, 0, 0);
 glRotatef(180, 0, 1, 0); 
 glDisable( GL_LIGHTING );
 gluSphere(QuadList,0.48,14,14);
 glEnable( GL_LIGHTING );
 glPopMatrix();
 glPushMatrix();
 glPopMatrix();
}


//Golems use a cylinder between the two end pieces.
void PaintGolem(ColourType Colour)
{
 PaintObject(GolemEndsList, Tgolemends.texID, GL_QUADS, ISGREY, false, 1.0);
 if(ColorCode == 1) 
 {
   glDisable( GL_TEXTURE_2D);
   glColor3ub(PGOLEM, PGOLEM, PGOLEM);
 }
 else
 {
  glEnable( GL_TEXTURE_2D );
  glBindTexture( GL_TEXTURE_2D, Tgolem.texID);     
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glColor3f(0.7, 0.7, 0.7);
  glShadeModel(GL_SMOOTH);
 }
 glPushMatrix();
 glTranslatef(0.5, 0.05, -0.5); 
 glRotatef(-90, 1, 0, 0); 
 gluCylinder(QuadList, 0.48, 0.40, 0.9, 10, 10);
 glPopMatrix();
}

//Paints the winning screen.  The functions are pretty much self-explanatory from here.
void PaintWinner()
{
 glPushMatrix();
 glEnable( GL_TEXTURE_2D );     
 glBindTexture( GL_TEXTURE_2D, Tsuccess.texID);      
 glEnable( GL_BLEND ); 
 glColor4f(1, 1, 1, 1);
 glBegin(GL_QUADS);
   glTexCoord2d(0,0);
   glVertex3f(60, 30, 0.9999);                
   glTexCoord2d(0,1);   
   glVertex3f(740, 30, 0.9999);     
   glTexCoord2d(1,1);   
   glVertex3f(740, 200, 0.9999);     
   glTexCoord2d(1,0);  
   glVertex3f(60, 200, 0.9999);     
 glEnd();
 glPopMatrix();
 PaintGameMenu();        
}

void PaintGameMenu()
{     
 glPushMatrix();
 glEnable( GL_TEXTURE_2D );     
 glBindTexture( GL_TEXTURE_2D, Tgmenu.texID);      
 glEnable( GL_BLEND ); 
 glColor4f(1, 1, 1, 1);
 glBegin(GL_QUADS);
   glTexCoord2d(0,0);
   glVertex3f(160, 230, 0.9999);                
   glTexCoord2d(1,0);   
   glVertex3f(660, 230, 0.9999);     
   glTexCoord2d(1,1);   
   glVertex3f(660, 370, 0.9999);     
   glTexCoord2d(0,1);  
   glVertex3f(160, 370, 0.9999);     
 glEnd();
 glPopMatrix();  
 PaintHelp(0);         
}


void PaintLoser()
{
 glPushMatrix();
 glEnable( GL_TEXTURE_2D );     
 glBindTexture( GL_TEXTURE_2D, Tfailure.texID);      
 glEnable( GL_BLEND ); 
 glColor4f(1, 1, 1, 1);
 glBegin(GL_QUADS);
   glTexCoord2d(0,0);
   glVertex3f(60, 30, 0.9999);                
   glTexCoord2d(0,1);   
   glVertex3f(740, 30, 0.9999);     
   glTexCoord2d(1,1);   
   glVertex3f(740, 200, 0.9999);     
   glTexCoord2d(1,0);  
   glVertex3f(60, 200, 0.9999);     
 glEnd();
 glPopMatrix();
 PaintGameMenu();        
}


void PaintBackground()
{
 glPushMatrix();
 glTranslatef(-49, -37.5, -90);
 PaintObject(BackgroundList, Tbackg.texID, GL_QUADS, ISGREY, true, 1.0);
 glPopMatrix();     
}

//This paints a string.  It'll actually account for newlines, making me pretty happy.
//I did make a few changes, but it's actually the paintstring function which I got
//from the OpenGL Lighthouse tutorial.
void PaintString(double x, double y, void* font, char* string)
{  
 if(!StillShifting())
 {
  char* c;
  int times;
  if(font == GLUT_BITMAP_HELVETICA_18 || font == GLUT_BITMAP_TIMES_ROMAN_24) times = 2;
  else times = 1;
  int SetWhite = 0;
  glDisable(GL_BLEND);
  glDisable(GL_TEXTURE_2D);
  glDisable(GL_LIGHTING);

  while(times > 0)
  {
   if(times > 1) glColor3f(1.0,1.0,1.0);
   else glColor3f(0.0,0.0,0.0);
   
   if(times > 1) glRasterPos3f(x-2,y-2,0.9999);
   else glRasterPos3f(x, y, 0.9999);

   for (c = string; *c != '\0'; c++) 
   {      
    if(*c == '_' && font != GLUT_BITMAP_TIMES_ROMAN_24) *c = ' ';
     glutBitmapCharacter(font, *c);
     if(*c == '\n') 
     {
      y = y+15;
      glRasterPos3f(x,y,0.9999); //Relax.  The bigger fonts don't use a \n, so this is fine.
     }
   }
   times--;
  }
  glEnable(GL_LIGHTING);
 }
}

void PaintWorld(bool Anim)
{
 AnimatePidIcon(Anim);     

 if(ShiftX > 1) ShiftX -= 2; 
 else if(ShiftX < -1) ShiftX += 2;
 else if(ShiftX >= -1 && ShiftX <= 1) { Shifting = false; ShiftX = 0;}
 glPushMatrix(); 
  glEnable( GL_TEXTURE_2D );     
  glEnable( GL_BLEND ); 
  
  if(StillShifting())
  {
   glBindTexture( GL_TEXTURE_2D, Tlastg.texID); 
   glColor4f(1, 1, 1, 1);
   glBegin(GL_QUADS);
    glTexCoord2d(0,1);
    glVertex3f(0, 0, 0.9997);    
    glTexCoord2d(1,1);   
    glVertex3f(800, 0, 0.9997);     
    glTexCoord2d(1,0);   
    glVertex3f(800, 568, 0.9997);     
    glTexCoord2d(0,0);  
    glVertex3f(0, 568, 0.9997);     
   glEnd();     
  }
  
  glBindTexture( GL_TEXTURE_2D, Tmcurrent);      
  glColor4f(1, 1, 1, 1);
  glBegin(GL_QUADS);
   glTexCoord2d(0,1);
   glVertex3f(0 + ShiftX, 0, 0.9998);    
   glTexCoord2d(1,1);   
   glVertex3f(800 + ShiftX, 0, 0.9998);     
   glTexCoord2d(1,0);   
   glVertex3f(800 + ShiftX, 568, 0.9998);     
   glTexCoord2d(0,0);  
   glVertex3f(0 + ShiftX, 568, 0.9998);     
  glEnd();  
  
  glBindTexture( GL_TEXTURE_2D, Tworldhud.texID );
  glBegin(GL_QUADS);
   glTexCoord2d(0,1);
   glVertex3f(0, 568, 0.9998);    
   glTexCoord2d(1,1);   
   glVertex3f(800, 568, 0.9998);     
   glTexCoord2d(1,0);   
   glVertex3f(800, 600, 0.9998);     
   glTexCoord2d(0,0);  
   glVertex3f(0, 600, 0.9998);     
  glEnd();  
  int NumMoves = GetLevelByNumMoves();
  if(NumMoves < 0) glBindTexture( GL_TEXTURE_2D, Twstarno.texID );
  else glBindTexture( GL_TEXTURE_2D, Twstaryes.texID );
  glBegin(GL_QUADS);
   glTexCoord2d(0,1);
   glVertex3f(342, 568, 0.9999);    
   glTexCoord2d(1,1);   
   glVertex3f(374, 568, 0.9999);     
   glTexCoord2d(1,0);   
   glVertex3f(374, 600, 0.9999);     
   glTexCoord2d(0,0);  
   glVertex3f(342, 600, 0.9999);     
  glEnd();  
 glPopMatrix(); 

 if(NumMoves < 0)  PaintString(395, 589, GLUT_BITMAP_HELVETICA_12, "No Solution Exists");   
 else
 {
   char Moving[12];
   char moves[3];
   itoa(NumMoves, moves, 10);
   strcpy(Moving, moves);
   strcat(Moving, " Moves");
   PaintString(380, 589, GLUT_BITMAP_HELVETICA_12, Moving); 
   PaintString(450, 589, GLUT_BITMAP_HELVETICA_12, "(V)iew Solution"); 
 }
 
 char TComplete[4];
 char TMoves[5];
 itoa(GetTotalComplete(), TComplete, 10); 
 itoa(GetTotalMoves(), TMoves, 10); 
 char Display[40];
 strcpy(Display, GetPlayerName());
 strcat(Display, " - ");
 strcat(Display, TComplete);
 strcat(Display, " levels in ");
 strcat(Display, TMoves);
 strcat(Display, " moves!");
 PaintString(550, 589, GLUT_BITMAP_HELVETICA_12, Display);   
 PaintString(15, 589, GLUT_BITMAP_HELVETICA_12, GetWorldName());
 PaintString(178, 589, GLUT_BITMAP_HELVETICA_12, "Level: ");
 PaintString(215, 589, GLUT_BITMAP_HELVETICA_12, GetLevelName());   
 
//Paint the location of each level with a line running up or down and a line running right or left
//to the next location.
 if(!StillShifting())
 {
  double FirstX = GetWorldX();
  double FirstY = GetWorldY();

  double NewX = FirstX;
  double NewY = FirstY;
  do
  {
   glDisable( GL_TEXTURE_2D );   
   glDisable( GL_LIGHTING );  
   glPushMatrix();
   glTranslatef(NewX-2, NewY-2, 0);
   glColor3f(0.0,0.0,0.0);
   glBegin(GL_QUADS); 
    glVertex3f(0, 0, 0.99987);                
    glVertex3f(36, 0, 0.99987);     
    glVertex3f(36, 36, 0.99987);     
    glVertex3f(0, 36, 0.99987);        
   glEnd();  
   glColor3f(1.0,1.0,1.0);  //Draw a big black box with a smaller white box.
   if(GetLevelByID() == 0) { glEnable( GL_TEXTURE_2D ); glBindTexture( GL_TEXTURE_2D, Tlblinking[0].texID ); }
   glTranslatef(2,2,0);     //These go by the co-ordinates provided by the WLD file.
   glBegin(GL_QUADS); 
    glTexCoord2d(0,0);
    glVertex3f(0, 0, 0.9999);    
    glTexCoord2d(1,0);            
    glVertex3f(32, 0, 0.9999);     
    glTexCoord2d(1,1);
    glVertex3f(32, 32, 0.9999);     
    glTexCoord2d(0,1);
    glVertex3f(0, 32, 0.9999);        
   glEnd();   
   glDisable( GL_TEXTURE_2D ); 
   glPopMatrix();
   double OldX = NewX;   //Now I'm gonna calculate and draw the line to the next level.
   double OldY = NewY;
   IncreaseMenuOption();  
   NewX = GetWorldX();
   NewY = GetWorldY();
   int Pushit = 0; int Pushwhite = 0;
   if(NewX < OldX){ Pushit = -7; Pushwhite = -2; }
  
   glPushMatrix();
   glColor3f(0.0,0.0,0.0);
   glBegin(GL_QUADS); 
     glVertex3f(OldX+13, OldY+13, 0.99985);      //Right or left.          
     glVertex3f(NewX+19 + Pushit, OldY+13, 0.99985);     
     glVertex3f(NewX+19 + Pushit, OldY+20, 0.99985);     
     glVertex3f(OldX+13, OldY+20, 0.99985);        
 
     glVertex3f(NewX+12, OldY+19, 0.99985);      //Up or down.     
     glVertex3f(NewX+12, NewY+13, 0.99985);     
     glVertex3f(NewX+19, NewY+13, 0.99985);     
     glVertex3f(NewX+19, OldY+19, 0.99985);            
   glEnd();  
   glColor3f(1.0,1.0,1.0);  //Do the same path in white.
   glBegin(GL_QUADS); 
     glVertex3f(OldX+15, OldY+15, 0.99986);      //Right or left.          
     glVertex3f(NewX+17 + Pushwhite, OldY+15, 0.99986);     
     glVertex3f(NewX+17 + Pushwhite, OldY+17, 0.99986);     
     glVertex3f(OldX+15, OldY+17, 0.99986);        
 
     glVertex3f(NewX+15, OldY+17, 0.99986);      //Up or down.     
     glVertex3f(NewX+15, NewY+15, 0.99986);     
     glVertex3f(NewX+17, NewY+15, 0.99986);     
     glVertex3f(NewX+17, OldY+17, 0.99986);            
   glEnd();    
   glPopMatrix();    
  } while(NewX != FirstX || NewY != FirstY);
 
  //Paint the penguin icon.
  glEnable( GL_TEXTURE_2D );
  glEnable( GL_LIGHTING );
  glPushMatrix();
  glTranslatef(GetWorldX(), GetWorldY(), 0);
  if(GetLevelByID() != 0) { glBindTexture( GL_TEXTURE_2D, Tlpid[PidIconAnimate].texID); }
  else { glBindTexture( GL_TEXTURE_2D, Tlblinking[(PidIconAnimate % 2)].texID); }
  glBegin(GL_QUADS);
   glTexCoord2d(0,0);
   glVertex3f(0, 0, 1.0);                
   glTexCoord2d(1,0);   
   glVertex3f(32, 0, 1.0);     
   glTexCoord2d(1,1);   
   glVertex3f(32, 32, 1.0);     
   glTexCoord2d(0,1);  
   glVertex3f(0, 32, 1.0);     
  glEnd();
  glPopMatrix();
 }
}

void PaintSelections()
{
 char* ScanSelection = GetSaveInfo();
 int NumMoves[4];
 int NumLevels[4];
 char* Name[4];
 Name[0] = (char*) malloc(128); 
 Name[1] = (char*) malloc(128); 
 Name[2] = (char*) malloc(128); 
 Name[3] = (char*) malloc(128);    

 sscanf(ScanSelection, "%d %d %s %d %d %s %d %d %s %d %d %s %d %d %s",
 &NumLevels[0], &NumMoves[0], Name[0],  &NumLevels[1], &NumMoves[1], Name[1],
 &NumLevels[2], &NumMoves[2], Name[2],  &NumLevels[3], &NumMoves[3], Name[3]);
 int y = 0;
 for(int i = 0; i < 4; i++)
 {
  char TLevels[4];
  char TMoves[5]; 
  itoa(NumLevels[i], TLevels, 10); 
  itoa(NumMoves[i], TMoves, 10);  
  if(strcmp(Name[i], "<nothing>") != 0) 
  {
   PaintString(180, 190+y, GLUT_BITMAP_TIMES_ROMAN_24, Name[i]);   
   PaintString(430, 160+y, GLUT_BITMAP_TIMES_ROMAN_24, TMoves);   
   PaintString(400, 190+y, GLUT_BITMAP_TIMES_ROMAN_24, "Total Moves");       
   PaintString(660, 160+y, GLUT_BITMAP_TIMES_ROMAN_24, TLevels);     
   PaintString(650, 190+y, GLUT_BITMAP_TIMES_ROMAN_24, "Levels");      
  }
  else PaintString(180, 190+y, GLUT_BITMAP_TIMES_ROMAN_24, "Start a New Game");   
  free(Name[i]); 
  y += 120;
 }
 if(Confirm() && !isNewGame())
 {
  PaintString(250, 55, GLUT_BITMAP_TIMES_ROMAN_24, "Continue");
  PaintString(500, 55, GLUT_BITMAP_TIMES_ROMAN_24, "Delete");                      
 }
 else if(Confirm())
 { 
  PaintString(190, 55, GLUT_BITMAP_TIMES_ROMAN_24, "Enter Your Name: ");
  PaintString(400, 55, GLUT_BITMAP_TIMES_ROMAN_24, GetNewName());  
 }
}

//Okay, this is admittedly kinda messy.  It retrieves the current screen from
//Menu.h and sets the current screen appropriately.
void PaintTitle(bool Anim)
{
 if(GetScreen() == STITLE) { Tmcurrent = Ttitle.texID; }
 else if(GetScreen() == SHELP1) Tmcurrent = Thelp1.texID;
 else if(GetScreen() == SHELP2) Tmcurrent = Thelp2.texID;
 else if(GetScreen() == SHELP3) Tmcurrent = Thelp3.texID;  
 else if(GetScreen() == SWORLD) Tmcurrent = Tbackg.texID;
 else if(GetScreen() == SSAVE) Tmcurrent = Tsave.texID;

 PaintString(5, 15, GLUT_BITMAP_HELVETICA_12, "(F)ullscreen  (P)lay Sound");       
 if(GetScreen() == SWORLD) //If we're drawing the world map, then display
   {                         //some 2D info to the user.           
    PaintWorld(Anim);
   }
 else
 {
  glPushMatrix();
  glEnable( GL_TEXTURE_2D );     
  glBindTexture( GL_TEXTURE_2D, Tmcurrent);      
  glEnable( GL_BLEND ); 
  glColor4f(1, 1, 1, 1);
  glBegin(GL_QUADS);
   glTexCoord2d(0,0);
   glVertex3f(0, 0, 0.9998);                
   glTexCoord2d(1,0);   
   glVertex3f(1024, 0, 0.9998);     
   glTexCoord2d(1,1);   
   glVertex3f(1024, 1024, 0.9998);     
   glTexCoord2d(0,1);  
   glVertex3f(0, 1024, 0.9998);     
  glEnd();
  glPopMatrix();
 }
 
 if(GetScreen() == STITLE || GetScreen() == SSAVE)
 { 
  if(Confirm())
  {
   glPushMatrix();
   glDisable( GL_TEXTURE_2D );     
   glEnable( GL_BLEND ); 
   glColor4f(0.0, 0.3, 0.6, 0.5);
   glBegin(GL_QUADS);
    glTexCoord2d(0,0);
    glVertex3f(180, 20, 0.99985);                
    glTexCoord2d(1,0);   
    glVertex3f(660, 20, 0.99985);     
    glTexCoord2d(1,1);   
    glVertex3f(660, 70, 0.99985);     
    glTexCoord2d(0,1);  
    glVertex3f(180, 70, 0.99985);     
   glEnd();
   glPopMatrix();  
  }
  
  if(!(Confirm() && isNewGame()))
  {
   glEnable(GL_BLEND);
   glEnable(GL_TEXTURE_2D);
   glColor4f(1,1,1,1);                
   glPushMatrix();
   glTranslatef(GetMenuX(Anim), GetMenuY(Anim), 0);
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); 
   glBindTexture( GL_TEXTURE_2D, Tlice.texID);
   glBegin(GL_QUADS);
    glTexCoord2d(0,0);
    glVertex3f(0, 0, 1.0);                
    glTexCoord2d(1,0);   
    glVertex3f(32, 0, 1.0);     
    glTexCoord2d(1,1);   
    glVertex3f(32, 32, 1.0);     
    glTexCoord2d(0,1);  
    glVertex3f(0, 32, 1.0);     
   glEnd();
   glPopMatrix();
  }
 }
 if(GetScreen() == SSAVE)
 {
  PaintSelections();     
 }
}

//Puts together and displays the total HUD.
void PaintHUD()
{
 glPushMatrix();
 glEnable( GL_TEXTURE_2D );     
 glBindTexture( GL_TEXTURE_2D, Thud.texID);      
 glEnable( GL_BLEND ); 
 glColor4f(1, 1, 1, 1);
 glBegin(GL_QUADS);
   glTexCoord2d(0,0);
   glVertex3f(100, 472, 0.9998);                
   glTexCoord2d(1,0);   
   glVertex3f(1124, 472, 0.9998);     
   glTexCoord2d(1,1);   
   glVertex3f(1124, 600, 0.9998);     
   glTexCoord2d(0,1);  
   glVertex3f(100, 600, 0.9998);     
 glEnd();

 glBindTexture( GL_TEXTURE_2D, Tlcurrent);
 glColor4f(1, 1, 1, 1);
 glBegin(GL_QUADS);
   glTexCoord2d(1,1);
   glVertex3f(377, 511, 0.99985);                
   glTexCoord2d(0,1);   
   glVertex3f(409, 511, 0.99985);     
   glTexCoord2d(0,0);   
   glVertex3f(409, 543, 0.99985);     
   glTexCoord2d(1,0);  
   glVertex3f(377, 543, 0.99985);     
 glEnd(); 
 
 glDisable(GL_TEXTURE_2D);
 if(Active == ISRED) glColor4f(1,0,0,1);
 else glColor4f(1,1,0,1);

 glBegin(GL_QUADS);
   glTexCoord2d(1,1);
   glVertex3f(366, 552, 0.99985);                
   glTexCoord2d(0,1);   
   glVertex3f(410, 552, 0.99985);     
   glTexCoord2d(0,0);   
   glVertex3f(410, 556, 0.99985);     
   glTexCoord2d(1,0);  
   glVertex3f(366, 556, 0.99985);     
 glEnd();  
 
 glPopMatrix();
}

void PaintCursor(int x, int y)
{
 glPushMatrix();
 glEnable( GL_TEXTURE_2D );     
 glBindTexture( GL_TEXTURE_2D, Tcursor.texID);      
 glEnable( GL_BLEND ); 
 glColor4f(1, 1, 1, 1);
 glBegin(GL_QUADS);
   glTexCoord2d(0,0);
   glVertex3f(x-10, y-1, 1);                
   glTexCoord2d(0,1);   
   glVertex3f(x+22, y-1, 1);     
   glTexCoord2d(1,1);   
   glVertex3f(x+22, y+31, 1);     
   glTexCoord2d(1,0);  
   glVertex3f(x-10, y+31, 1);     
 glEnd();
 glPopMatrix();
}

//Sets me up for picking based on color codes.
void StartPicking()
{
 ColorCode = 1;    
}

//Stops it.
void StopPicking()
{
 ColorCode = 0;    
}

//We set the new active colour in the level.
void SetActiveColour(ColourType newActive)
{
 Active = newActive;
}

void SetSkidmark()
{
 Skidmark = true;    
}

void PaintMapBottom()
{
 glPushMatrix();
 glTranslatef(0.0, 0.1, 0.0);
  PaintObject(WallBoundsList, Tbottom.texID, GL_QUADS, ISGREY, false, 1.0); 
  PaintObject(SkyList, Tsky.texID, GL_QUADS, ISGREY, false, 1.0);
 glPopMatrix();
}

void PaintHole(int x, int y, bool North, bool East, bool South, bool West)
{
 glPushMatrix();
   glTranslatef(x, 0, y);     
   PaintObject(HoleList, Thole.texID, GL_QUADS, ISGREY, false, 1.0);

   glEnable( GL_TEXTURE_2D );     
   glBindTexture( GL_TEXTURE_2D, Tholewall.texID);      
   glEnable( GL_BLEND ); 
   glColor4f(1, 1, 1, 1);
   glBegin(GL_QUADS);
   if(North)
   {
      glTexCoord2d(0,0);
      glVertex3f(0.0, 0.1, -0.999);                
      glTexCoord2d(1,0);   
      glVertex3f(1.0, 0.1, -0.999);     
      glTexCoord2d(1,1);   
      glVertex3f(1.0, -0.6, -0.999);     
      glTexCoord2d(0,1);  
      glVertex3f(0.0, -0.6, -0.999);
   }
   if(South)
   {
      glTexCoord2d(0,0);
      glVertex3f(0.0, 0.1, 0.001);                
      glTexCoord2d(1,0);   
      glVertex3f(0.999, 0.1, 0.001);     
      glTexCoord2d(1,1);   
      glVertex3f(0.999, -0.6, 0.001);     
      glTexCoord2d(0,1);  
      glVertex3f(0.0, -0.6, 0.001);
   }
   if(West)
   {
      glTexCoord2d(0,0);
      glVertex3f(0.001, 0.1, 0.001);                
      glTexCoord2d(0,1);   
      glVertex3f(0.001, -0.6, 0.001);     
      glTexCoord2d(1,1);   
      glVertex3f(0.001, -0.6, -0.999);     
      glTexCoord2d(1,0);  
      glVertex3f(0.001, 0.1, -0.999);
   }
   if(East)
   {
      glTexCoord2d(0,0);
      glVertex3f(0.999, 0.1, 0.001);                
      glTexCoord2d(0,1);   
      glVertex3f(0.999, -0.6, 0.001);     
      glTexCoord2d(1,1);   
      glVertex3f(0.999, -0.6, -0.999);     
      glTexCoord2d(1,0);  
      glVertex3f(0.999, 0.1, -0.999);
   }         
   glEnd();
   
 glPopMatrix();      
}

//This function paints the holes in the floor.  It calculates holes connecting together
//and places walls accordingly.
void PaintHoles(Level theLevel)
{
  for(int i = 0; i < theLevel.Size; i++)
   for(int j = 0; j < theLevel.Size; j++)
   {
    if(theLevel.Walls[i][j].Hole)  //Check adjacent holes.  
    {
     bool North = false;
     bool West = false;  //These will be set to true if walls need to be placed there.
     bool East = false;  //ie. the pit ends at this spot for one of four directions.
     bool South = false;
     if(i-1 < 0 || !theLevel.Walls[i-1][j].Hole) West = true;
     if(i+1 >= theLevel.Size || !theLevel.Walls[i+1][j].Hole) East = true;
     if(j-1 < 0 || !theLevel.Walls[i][j-1].Hole) North = true;
     if(j+1 >= theLevel.Size || !theLevel.Walls[i][j+1].Hole) South = true;     
     PaintHole(i,j, North, East, South, West);
    }
   }
}

//This function paints the walls for a tile at the appropriate time as called by the functions
//below.
void PaintWalls(WallTile theWall)
{
   GLuint WallTexture = Twall.texID;
   if(theWall.Colour != ISGREY) WallTexture = Tgwall.texID;
   //Now we paint the appropriate walls with the appropriate colour.
   if(theWall.Blocks[0]) PaintObject(WallNorthList, WallTexture, GL_QUADS, theWall.Colour, false, 1.0);
   if(theWall.Blocks[1]) 
   {
    glPushMatrix();
    glTranslatef(0,0,-1);
    glRotatef(-90,0,1,0);    
    PaintObject(WallNorthList, WallTexture, GL_QUADS, theWall.Colour, false, 1.0);
    glPopMatrix();
   } 
   if(theWall.Blocks[2]) 
   {
     glPushMatrix();
     glTranslatef(0,0,0.97);
     PaintObject(WallNorthList, WallTexture, GL_QUADS, theWall.Colour, false, 1.0); 
     glPopMatrix();
   }
   if(theWall.Blocks[3])
   {
    glPushMatrix();
    glTranslatef(1,0,0);
    glRotatef(90,0,1,0); 
    PaintObject(WallNorthList, WallTexture, GL_QUADS, theWall.Colour, false, 1.0);
    glPopMatrix();
   }    
}

//Paints see-through walls.
void PaintInvisibleWalls(Level theLevel)
{
  for(int i = 0; i < theLevel.Size; i++)
   for(int j = 0; j < theLevel.Size; j++)
   {
    if(theLevel.Walls[i][j].Colour != ISGREY && theLevel.Walls[i][j].Colour != Active) 
    {
     glPushMatrix();
     glTranslatef(i, 0, j);                                      
     glScalef(1,0.7,1);
     PaintWalls(theLevel.Walls[i][j]);
     glPopMatrix();
    }
   }
}

//This function paints the floor tiles over the reflective surface, be it dirt or the mirrory effect.
void PaintFloor(WallTile theWall, int x, int y)
{
 glPushMatrix();
   glTranslatef(x, 0, y);     
   glTranslatef(0,0.01,0);
   if(theWall.Dirty) PaintObject(WallBottomList, Tdirt.texID, GL_QUADS, ISGREY, false, 1.0);    
   else if(!theWall.Hole) PaintObject(WallBottomList, Tmirror.texID, GL_QUADS, ISGREY, false, 1.0);    
   else PaintObject(HoleList, Thole.texID, GL_QUADS, ISGREY, false, 1.0);
 glPopMatrix();
}

//This function paints walls.
void PaintWall(WallTile theWall, int x, int y, bool Reflection)
{
 //If this is a reflection, don't paint the bottom to the level again, otherwise...that'd look weird.
 glPushMatrix();
 glTranslatef(x, 0, y);
 if(!theWall.Hole)  //If there is a hole, we just don't paint anything.
 {
  if(!Reflection)
  {
   glEnable( GL_BLEND );
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); 
   PaintObject(WallBottomList, Tbottom.texID, GL_QUADS, ISGREY, false, 0.8); //Paint the bottom. 
   
  }

  if(theWall.Colour == ISGREY || theWall.Colour == Active)
  {
   PaintWalls(theWall);
  }
 } 
 glPopMatrix();
}

//This paints the particles once everything else has been drawn.

void PaintParticles(bool Anim)
{
 PaintSystem(Anim); //Tell the particle system to draw itself.
}

//This paints the given tile.  So it takes some info from it and determines what and where
//to draw the object.

PlayTile PaintCharacter(PlayTile theTile, bool Anim)
{
 glPushMatrix();
 float AnimSpeed = 0.2; //This constant just moves the object a very little bit
                         //between frames.

 //If the tile should be moving in the Painter(cuz the X,Ys don't match up)
 //Note that we only animate the tile if we've been told to by the Driver.
 //This is to match up a constant framerate across all computers (hopefully).
 if(Anim)
 {
  if(theTile.paintX < theTile.X) 
  {
   theTile.paintX += AnimSpeed;
   if(theTile.paintX > theTile.X) theTile.paintX = theTile.X;
  }
  else if(theTile.paintX > theTile.X) 
  {
   theTile.paintX -= AnimSpeed;
   if(theTile.paintX < theTile.X) theTile.paintX = theTile.X;
  } 
  if(theTile.paintY < theTile.Y) 
  {
   theTile.paintY += AnimSpeed;
   if(theTile.paintY > theTile.Y) theTile.paintY = theTile.Y;
  }
  else if(theTile.paintY > theTile.Y) 
  {
    theTile.paintY -= AnimSpeed;
   if(theTile.paintY < theTile.Y) theTile.paintY = theTile.Y;
  } 
 }

//Then we centre it and scale it down a little.  This is so that
//the object will draw properly in-between walls.
 glTranslatef(theTile.paintX+0.04, 0.1, theTile.paintY-0.05);  
 glScalef(0.92, 0.92, 0.92);  
 //If my tile is 'alive', consider the direction it's facing.  Rotate
 //the object to face that direction.  To be honest, I could be doing
 //this with ALL of my objects...but only Heroes and Enemies have
 //appreciable point ats.
 if(theTile.Alignment == HERO || theTile.Alignment == ENEMY)
 {
  if(theTile.paintM == EAST)
  {
   glTranslatef(0.5, 0, 0.5);
   glRotatef(90, 0, 1, 0);
   glTranslatef(0.5, 0, 0.5);
  }
  if(theTile.paintM == WEST)
  {
   glTranslatef(0.5, 0, 0.5);
   glRotatef(-90, 0, 1, 0);
   glTranslatef(-1.5, 0, 0.5);
  }
  if(theTile.paintM == NORTH)
  {
   glTranslatef(0.5, 0, 0.5);
   glRotatef(180, 0, 1, 0);
   glTranslatef(-0.5, 0, 1.5);
  }    
 }
 
 //Of course, the PointDirection combines with the Face direction.  The code below
 //is held exclusively by arrows at the moment.  Also, an object will have either
 //one or the other and not both.  This also could change.
 
 if(theTile.PointDirection == PEAST)
 {
  glTranslatef(0.5, 0, 0.5);
  glRotatef(-90, 0, 1, 0); 
  glTranslatef(-1.5, 0, 0.5);
 }
 else if(theTile.PointDirection == PWEST)
 {
  glTranslatef(0.5, 0, 0.5);
  glRotatef(90, 0, 1, 0);  
  glTranslatef(0.5, 0, 0.5);     
 }
 else if(theTile.PointDirection == PSOUTH)
 {
  glTranslatef(0.5, 0, 0.5);
  glRotatef(180, 0, 1, 0);
  glTranslatef(-0.5, 0, 1.5);
 }   
 
 //If the object is an inactive colour, we make it smaller so that
 //when it turns on, it grows in size.  Well, I thought it looked neat anyway.
 if((theTile.Colour == ISRED && Active == ISYELLOW) ||
    (theTile.Colour == ISYELLOW && Active == ISRED))
 {
  glScalef(0.7, 0.7, 0.7);
  glTranslatef(0.2, 0, -0.2);
 }
 
 //Use the String to determine which object to paint.
 if(strcmp("PID", theTile.Name) == 0) PaintPid(theTile.Colour, Anim);
 else if(strcmp("CRYSTAL", theTile.Name) == 0)
 {
  PaintCrystal(theTile.Colour, Anim);
  if(theTile.Colour == Active || theTile.Colour == ISGREY)
    RegisterSystem(theTile.paintX, theTile.paintY, 2);
  else
    RegisterSystem(theTile.paintX, theTile.paintY, 2);
 }  
 else if(strcmp("CLOSEDEXIT", theTile.Name) == 0) PaintExit(theTile.Colour);
 else if(strcmp("OPENEXIT", theTile.Name) == 0) 
 {
  PaintOpenExit(theTile.Colour); 
  RegisterSystem(theTile.paintX, theTile.paintY, 3);  
 }
 else if(strcmp("ALLY", theTile.Name) == 0) PaintAlly(theTile.Colour);
 else if(strcmp("GOLEM", theTile.Name) == 0) PaintGolem(theTile.Colour);
 else if(strcmp("FURY", theTile.Name) == 0) 
 {
  PaintFury(theTile.Colour);
  RegisterSystem(theTile.paintX, theTile.paintY, 1);
 }
 else if(strcmp("FLAME", theTile.Name) == 0) 
 {
  PaintFlame(theTile.Colour);
  RegisterSystem(theTile.paintX, theTile.paintY, 1);
 }
 else if(strcmp("SWITCH", theTile.Name) == 0) 
 {
   if(theTile.Colour == SETRED) PaintRedSwitch(theTile.Colour);
   else if(theTile.Colour == SETYELLOW) PaintYellowSwitch(theTile.Colour);
 }
 else if(strcmp("TRAP", theTile.Name) == 0) PaintTrap(theTile.Colour);
 else if(strcmp("ARROW", theTile.Name) == 0) PaintArrow(theTile.Colour);
 else if(strcmp("HVBLOCK", theTile.Name) == 0) PaintHVBlock(theTile.Colour);
 else if(strcmp("VBLOCK", theTile.Name) == 0) PaintVBlock(theTile.Colour);
 else if(strcmp("HBLOCK", theTile.Name) == 0) PaintHBlock(theTile.Colour);
 else if(strcmp("DBLOCK", theTile.Name) == 0) PaintDBlock(theTile.Colour);
 glPopMatrix(); 
 return theTile;
}

//This function paints the Info string.
void PaintInfo()
{
 PaintString(425, 514, GLUT_BITMAP_HELVETICA_10, Info);      
}

//This function sets the Info string depending on the object that was picked.
void PaintHelp(int Picked)
{
 if(Picked == 0)
 {
  Info = "";  
  Tlcurrent = Tlnothing.texID;
 }
 else if(Picked <= CPID)
 {
  Info = "The penguins are a noble breed.  They are\nworking together to find the crystals and to \nstop the invading forces of the Furies.\n\n But only Pid has the ability to control the ice, \nthe crystals, and the blocks.";     
  Tlcurrent = Tlpenguin.texID;
 }    
 else if(Picked <= CCRYSTAL)
 {
  Info = "One day, these mysterious gems appeared \nalong with the statues of stone and ice. \n\n Pid must find all of them whenever he can \nbefore he can leave the area!";     
  Tlcurrent = Tlcrystal.texID;  
 }
 else if(Picked <= CEXIT) 
 {
  Info = "These exits will lead penguins out of the area.\n\n  Pid must collect all the crystals before it \ncan be used.";    
  Tlcurrent = Tlexit.texID;
 }
 else if(Picked <= CGOLEM)
 {
  Info = "Always surrounding Pid, the golems of \nKouriku rise out of the ground to greet him. \n\n Silent statues which mark a dead era before \nthe need of crystals.  They will protect Pid \nfrom surrounding furies.";     
  Tlcurrent = Tlgolem.texID;
 }
 else if(Picked <= CFURY)
 {
  Info = "The Furies come from an uknown land.  \nThey are invading Freja and know that Pid, \nwith his powers of Kouriku summoning, is the \nonly one who can stand against them.  They \nwill burn him into ashes given the chance.";     
  Tlcurrent = Tlfury.texID;
 }
 else if(Picked <= CFLAME)
 {
  Info = "The Flame Generals of the Fury army are \nthe most devious of all.  Although all \nlittle Furies and other objects fall prey to \nPid's command of the ice, the Flames will \nmove in unexpected ways.";     
  Tlcurrent = Tlfury.texID;  
 } 
 else if(Picked <= CSWITCH)
 {
  Info = "With the powers of the Kouriku and the Furies \nclashing, areas have been pushed out of \nexistence.  These areas cannot be felt without \nthe reinforcing power of an Energy Switch. \n\n Pid, use your best judgment!";
  Tlcurrent = Tlswitch.texID;
 }
 else if(Picked <= CTRAP)
 {
  Info = "Planted by Flame Generals, these Trapmines \nare dangerous.  If a trapmine is touched by \nanything containing spirit, the trap will \nexplode.  However, Pid may be able to use \nthese to his advantage...";     
  Tlcurrent = Tltrap.texID;
 } 
 else if(Picked <= CARROW)
 {
  Info = "Arrows will redirect anything that comes near \nthem.  Using an arrow can surprise and \nconfuse your enemies.  If you want to go for a \nride, give these a shot!";
  Tlcurrent = Tlarrow.texID;
 }
 else if(Picked <= CBOX)
 {
  Info = "Boxes are constructs of the Penguin race in \norder to help Pid.  They aren't as easily \nmoved as Kouriku and only Pid can move \nthem.  They can be useful, or a hindrance. \n\n Hey...that black box isn't one of ours...";     
  Tlcurrent = Tlbox.texID;
 }
}

void ExitGraphics()
{   
  freeList(BoxList);
  freeList(BoxRimList);
  freeList(BackgroundList);
  freeList(MenuList);
  freeList(HUDList);   
  freeList(WallBottomList);   
  freeList(WallNorthList);     
  freeList(GolemEndsList);   
  freeList(PidFaceList);   
  freeList(PidBodyList);   
  freeList(PidFrontList);   
  freeList(PidFeetList);   
  freeList(TrapList);    
  freeList(TrapEdgeList);   
  freeList(SwitchBoxList);   
  freeList(SwitchList);   
  freeList(CrystalList);   
  freeList(ExitList);   
  freeList(ArrowTopList);   
  freeList(ArrowRimList);   
  freeList(WallBoundsList);
  freeList(HoleList);
  freeList(ArrowStemList);   
  freeList(SkyList);
  freeList(HoleList);
  gluDeleteQuadric(QuadList);    
}

