/*
KourikuGL
Michael Vendittelli 
*/

#ifndef gf_included_PlayTile_h
#define gf_included_PlayTile_h

#include "types.h"

struct aPlayTile
{
 char* Name;  //The name of the play tile.  This is used to identify it to the Painter.

 Direction Moving;  //The direction it is moving in, being Stopped, North, East, South, West.
 bool Moveable;  //If the tile can actually move or not.
 bool GivesDirections;  //If the tile gives directions to Blocks.
 bool ForceDirections;  //If the tile forces it's PointAt on other tiles.
 PointAt PointDirection;  //The direction in which the Tile is pointing.
 Direction paintM; //The direction in which the Tile is facing.  This is important only to the Painter. 
 PointAt NewMoving;  //The direction in which the tile will move after it's done moving this round.
 DirType MoveType;  //The type of movement, being it normal, reversed, or a Block.
 bool Allowed[4];  //What directions the tile may move in when given directions by either player or another tile.
 
 int X;            //Current X position on the board.
 int Y;            //Current Y position on the board.
 double paintX;    //Current X position relative to the Painter.
 double paintY;    //Current Y position relative to the Painter.
 ColourType Colour; //The colour of the Tile, be it Grey, Yellow, or Red.
 AlignType Alignment;  //The alignment of the Tile, be it Hero, Enemy, or Neutral.
 CollectType Collect;  //If the tile collects collectables, is a collectable, or is waiting on an open exit.
 bool Important;   //If this tile dies, player fails at the level.
 bool Murderer;    //If this tile murders anything liveable.  Opposite alignments automatically murder eachother.
 bool Suicidal;    //If this tile commits suicide after it kills.
 bool Liveable;    //If this tile is considered alive on the board.
 bool InTheWay;    //If the tile takes up space on the board.
 bool RemoveMe;    //If the tile is set to be removed in the current round because it died.
 bool Collected;   //We set this extra flag if it was removed, but collected instead of killed.
 bool Dirty;       //We set the dirty flag when a tile lands on dirt.
};

typedef struct aPlayTile PlayTile;

PlayTile MakePlayTile(char* pName, Direction pMoving, bool pMoveable, bool pGivesDirections, 
                   bool pForceDirections, PointAt pPointDirection, DirType pMoveType, 
                   bool pAllowed[4], int pX, int pY, ColourType pColour, AlignType pAlignment,
                   bool pMurderer, bool pSuicidal, bool pLiveable, bool pInTheWay, 
                   bool pImportant, CollectType pCollect);
//This function obviously makes a Tile as specified.  This function should normally be called
//by MakeStandardTile, unless the tile is custom-made outside of the standard editor.

PlayTile MakeStandardTile(char* pName, ColourType pColour, bool pMoveable, PointAt pPointDirection, int X, int Y);
//This makes one of the standard tiles as defined by the editor capabilities.

PlayTile SetDirectionByPlayer(PlayTile theTile, Direction NewDir);
//This sets the direction on the tile as specified by the player.  If this is called, it
//is highly likely that it will be eventually called on all tiles for the board.

bool Takeover(PlayTile TileA, PlayTile TileB, ColourType Active);
//true if TileA moves to the position of TileB.
//false if it should stay put.

int RemoveTile(PlayTile TileA, PlayTile TileB, ColourType Active);
//Tile A is hitting Tile B.
//Returns 0 if both are still present.  1 if removing TileA.  2 if removing TileB.  3 if
//removing both.

int DetermineNewDirection(PlayTile TileA, PlayTile TileB, ColourType Active);
//returns 0 if Tile A nor B are changing directions.  returns 1 if Tile A is forcing a change on
// TileB.  2 if tile B is forcing change on TileA.  3 if Tile A is giving directions to B.  

PlayTile SetDirectionByBoard(PlayTile theTile, Direction NewDir);

int SeeInteraction(PlayTile TileA, PlayTile TileB, ColourType Active);
//returns 0 if there is no possible interaction between these two tiles.
//returns 1 if there is an interaction.
//returns 2 if there could be an interaction, but TileB may be resolved beforehand.
#endif
