#include <al/alc.h>
#include <al/alut.c>

#include "Sound.h"
#include "ogg.h"

typedef struct 
{
 ALuint Buffer;
 ALuint Source;       
} FlatEffect;

ogg_stream ogg;
PlayList Songs[8];  //Where 0 is the menu music and 1 is the cutscene music.
int CurrentSong = 0;
int Started = 0;
int GameNumber = 2;

//Creating a separate source for each sound effect seems ridiculous as this would lend
//itself to the limitations of the channels.  I'll probably try to find a better sound library
//if I ever design a more complex game as this one seems to drop the ball in this one particular
//area.  I'm sure there's a more simple solution that I'm just not seeing.  Unfortunately, my
//other 'solutions' seem to cause memory leaks...not a good thing.

FlatEffect Bump;
FlatEffect Bye;
FlatEffect Collect;
FlatEffect Explode;
FlatEffect Hole;
FlatEffect Ski;
FlatEffect Menu;
FlatEffect Switch;
FlatEffect Success[5];
int NSuccess = 0;
FlatEffect Failure[5];
int NFailure = 0;

ALfloat SourcePos[] = { 0.0, 0.0, 0.0 };
ALfloat SourceVel[] = { 0.0, 0.0, 0.0 };
ALfloat ListenerPos[] = { 0.0, 0.0, 0.0 };
ALfloat ListenerVel[] = { 0.0, 0.0, 0.0 };
ALfloat ListenerOri[] = { 0.0, 0.0, -1.0,  0.0, 1.0, 0.0 };

ALboolean LoadALData(FlatEffect* Effect, char* File)
{
	// Variables to load into.

	ALenum format;
	ALsizei size;
	ALvoid* data;
	ALsizei freq;
	ALboolean loop;
	// Load wav data into a buffer.
	alGenBuffers(1, &Effect->Buffer);     	

	if(alGetError() != AL_NO_ERROR)
		return AL_FALSE;

	alutLoadWAVFile(File, &format, &data, &size, &freq, &loop);
	alBufferData(Effect->Buffer, format, data, size, freq);
	alutUnloadWAV(format, data, size, freq);

    alGenSources(1, &Effect->Source);

    alSourcei (Effect->Source, AL_BUFFER,   Effect->Buffer);
    alSourcef (Effect->Source, AL_PITCH,    1.0          );
    alSourcef (Effect->Source, AL_GAIN,     1.0          );
    alSourcefv(Effect->Source, AL_POSITION, SourcePos    );
    alSourcefv(Effect->Source, AL_VELOCITY, SourceVel    );
    alSourcei (Effect->Source, AL_LOOPING,  AL_FALSE     );

    if(alGetError() == AL_NO_ERROR)
		return AL_TRUE;

	return AL_FALSE;
}

void SetListenerValues()
{
	alListenerfv(AL_POSITION,    ListenerPos);
	alListenerfv(AL_VELOCITY,    ListenerVel);
	alListenerfv(AL_ORIENTATION, ListenerOri);
}

void PlayBye(){ alSourcePlay(Bye.Source);  }
void PlayBump(){ alSourcePlay(Bump.Source);  }
void PlayCollect(){ alSourcePlay(Collect.Source);  }
void PlayExplode(){ alSourcePlay(Explode.Source);  }
void PlayHole(){ alSourcePlay(Hole.Source);  }
void PlaySki(){ alSourcePlay(Ski.Source);  }
void PlayMenu(){ alSourcePlay(Menu.Source);  }
void PlaySwitch(){ alSourcePlay(Switch.Source);  }
void PlaySuccess()
{ 
     alSourcePlay(Success[NSuccess].Source);  
     NSuccess++;
     if(NSuccess > 4) NSuccess = 0;
}
void PlayFailure()
{ 
     alSourcePlay(Failure[NFailure].Source);  
     NFailure++;
     if(NFailure > 4) NFailure = 0;
}


void SetupSound()
{
 SetListenerValues(); 
 LoadALData(&Bye, "Sound/bye.wav");
 LoadALData(&Bump, "Sound/bump.wav");
 LoadALData(&Collect, "Sound/collect.wav");
 LoadALData(&Explode, "Sound/explode.wav");
 LoadALData(&Hole, "Sound/hole.wav");
 LoadALData(&Ski, "Sound/ski.wav"); 
 LoadALData(&Menu, "Sound/menu.wav");
 LoadALData(&Switch, "Sound/switch.wav");      

 LoadALData(&Success[0], "Sound/success1.wav");  
 LoadALData(&Success[1], "Sound/success2.wav");  
 LoadALData(&Success[2], "Sound/success3.wav");  
 LoadALData(&Success[3], "Sound/success4.wav");  
 LoadALData(&Success[4], "Sound/success5.wav");      

 LoadALData(&Failure[0], "Sound/failure1.wav");  
 LoadALData(&Failure[1], "Sound/failure2.wav");  
 LoadALData(&Failure[2], "Sound/failure3.wav");  
 LoadALData(&Failure[3], "Sound/failure4.wav");  
 LoadALData(&Failure[4], "Sound/failure5.wav");       
}

void SetupPlayList()
{
 alutInit(0, NULL);     
 
 Songs[0].File = "Sound/skankypad.ogg";    
 Songs[1].File = "Sound/whammy.ogg";    
 Songs[2].File = "Sound/deca.ogg";    
 Songs[3].File = "Sound/silicin.ogg";    
 Songs[4].File = "Sound/holdthephone.ogg";     
 Songs[5].File = "Sound/diamondfalls.ogg";     
 Songs[6].File = "Sound/proto.ogg";    
 Songs[7].File = "Sound/empty.ogg";      
}

void SoundBegin()
{
  ogg.open(Songs[CurrentSong].File);
  ogg.playback();  
}

void SoundExit()
{
 ogg.release();	
}

void PlayMusic()
{
 ogg.update();
 if(ogg.broken())
 {
  SoundExit();
  SoundBegin();                
 }
 if(!ogg.playing())
 {
  if(!ogg.playback())
   ogg.open(Songs[CurrentSong].File);
 }    
}

void ChooseSound(int value)
{
 if(Started) SoundExit();
 else Started = 1;
 if(value == 3) CurrentSong = 7;
 else if(value == 0) CurrentSong = 0;
 else if(value == 1) CurrentSong = 1;
 else if(value == 2)
 {
   CurrentSong = GameNumber;
   GameNumber++;
   if(GameNumber > 6) GameNumber=2;
 }
 SoundBegin();     
}

void KillSound()
{
 alDeleteBuffers(1, &Bye.Buffer);
 alDeleteSources(1, &Bye.Source);

 alDeleteBuffers(1, &Bump.Buffer);
 alDeleteSources(1, &Bump.Source);
 
 alDeleteBuffers(1, &Collect.Buffer);
 alDeleteSources(1, &Collect.Source);
 
 alDeleteBuffers(1, &Explode.Buffer);
 alDeleteSources(1, &Explode.Source);
 
 alDeleteBuffers(1, &Hole.Buffer);
 alDeleteSources(1, &Hole.Source);  

 alDeleteBuffers(1, &Menu.Buffer);
 alDeleteSources(1, &Menu.Source);  

 alDeleteBuffers(1, &Switch.Buffer);
 alDeleteSources(1, &Switch.Source);    

 for(int i = 0; i < 5; i++)
  {
  alDeleteBuffers(1, &Success[i].Buffer);
  alDeleteSources(1, &Success[i].Source);    
  alDeleteBuffers(1, &Failure[i].Buffer);
  alDeleteSources(1, &Failure[i].Source);     
 }

 alutExit();
}


